unit Helios_Shader_System;
{
Unit: Helios_Shader_System
By: Stuart Carey
Based On: Shaders.pas by Harry Painter
}

{$I Global_Conditionals.INC}

interface

uses StdCtrls, sysutils, classes, OpenGL15, Hashes, math3d{$IfDef HELIOS_APP},Helios_Types{$Else},Helios_Basic_Types{$EndIf};

const
  Rad = Pi / 180.0;
  StaticGamma = 5; // Used when not linked to helios

var
skyboxId : integer = -1;

type
enSortKey = (SK_NONE = 0, SK_PORTAL = 1, SK_SKY = 2, SK_OPAQUE = 3, SK_BANNER = 6, SK_TRANS = 7, SK_UNDERWATER = 8, SK_ADDITIVE = 9, SK_NEAREST = 16);
enTCGen =   (TCGEN_STANDARD = 0, TCGEN_ENVIRONMENT = 1, TCGEN_BASE = 2, TCGEN_LIGHTMAP = 3, TCGEN_VECTOR = 4);
enrgbGen =  (RGBGEN_IDENTITYLIGHTNING = 0, RGBGEN_IDENTITY = 1, RGBGEN_WAVE = 2, RGBGEN_ENTITY = 3, RGBGEN_ONEMINUSENTITY = 4,
                  RGBGEN_VERTEX = 5, RGBGEN_ONEMINUSVERTEX = 6, RGBGEN_LIGHTNINGDIFFUSE = 7);
entcmod =   (TCMOD_UNUSED = 0, TCMOD_ROTATE = 1, TCMOD_SCALE = 2, TCMOD_SCROLL = 3, TCMOD_STRETCH = 4, TCMOD_TRANSFORM = 5, TCMOD_TURB = 6);
enWave =    (WF_NONE = 0, WF_SIN = 1, WF_TRIANGLE = 2, WF_SQUARE = 3, WF_SAWTOOTH = 4, WF_INVSAWTOOTH = 5, WF_NOISE = 6);
enDeformType = (DEFORMVERTEXES_NONE = 0,	      // 0 parameters
                     DEFORMVERTEXES_WAVE = 1,	      // 6 parameters
                     DEFORMVERTEXES_NORMAL= 2,	      // 5 parameters
                     DEFORMVERTEXES_BULGE = 3,	      // 3 parameters
                     DEFORMVERTEXES_MOVE = 4,	      // 8 parameters
                     DEFORMVERTEXES_AUTOSPRITE	= 5,  // 0 parameters
                     DEFORMVERTEXES_AUTOSPRITE2	= 6); // 0 parameters

TDeformVertexes = record
  deformType : enDeformType;
  values : array[0..7] of single;
end;

TWaveFunc = record
  func : enWave;
  base : Single;
  amp  : Single;
  phase : Single;
  freq : Single;
  // state info
  curStatus : single;
  Direction : boolean;
end;

TrgbGen = record
  rgbGType :  enrgbGen;
  Values : array[0..5] of single;
  Wave : TWaveFunc;
end;

Ttcmod = record
  modType : entcmod;
  Values : array[0..5] of single;
  Wave : TWaveFunc;
end;

TFogParam = record
  Red : single;
  Green : single;
  Blue : single;
  Alpha : single;
  Distance : single;
end;

TLayer = record
  TextureNames : array of string;
  TextureId : array[0..7] of GLUINT;
  numOfTextures : byte;
  tcMods : array  of Ttcmod;
  numOftcMods : integer;
  rgbGen : array of TrgbGen;
  numOfrgbGen : integer;
  animFreq : Single;
  animFrameTime : integer; // 1 frame each animFreq ms
  doBlending : Boolean;
  doAlphaTest : Boolean;
  DepthWrite :Boolean;
  DepthFunc : integer;
  DestBlend : integer;
  SrcBlend : integer;
  AlphaFunc : integer;
  AlphaRef : single;
  alphaWave : TWaveFunc;
  UseLightmap : boolean;
  UseVertexColors : boolean;

  // state infos
  curAnimStat : integer;
  curTexture : integer;
  curColor : array[0..3] of byte; // rgba
  TexClamp :  Boolean;
  tcGen : enTcGen;

  //Default Scaling-value
  tcModScaleX  : single;
  tcModScaleY  : single;
  tcModStretchX : single;
  tcModStretchY : single;
  rgbFactor    : single;
  XOffset : single;
  YOffset : single;
  rotate : single;

  DynamicTexture : Boolean; // If true use the dynamic texture insted of any map the layer has!
  Draw : boolean;
  IsBumpMap : Boolean; // Blend can hold bumpmap or diffusemap
  IsDiffuseMap : Boolean;
end;

TShader = class(TCollectionItem)
  public
    ShaderFile : string;
    Name : string;
    lastTime : longword;
    Cull : integer;
    Source : string;
    Surface : cardinal;
    hsurface : cardinal;      // Stores Surfaces Like Ladder In Helios's Definition
    Contents : cardinal;
    Layers : array of TLayer;
    numOfLayers : integer;
    numOfDeformVertexes : integer;
    qerImage : string;
    qerId : GLUINT;
    NoMipMap : boolean;
    NoPicMip : boolean;
    IsSkyBox : boolean;
    effectIndex : integer;
    FogValues : array[0..1] of single;
    DeformVertexes : array of TDeformVertexes;
    Fog : TFogParam;
    hasFog : boolean;
    SortKey : enSortKey;
    TotallyTrans : boolean;
    NonOpaque : Boolean;
    SkyboxType : integer;
    SkyboxTextureNames : array of string;
    SkyboxTextureIds : array of cardinal;
    PolygonOffset : Boolean;

    //Doom 3
    BumpMap,SpecularMap,DiffuseMap : String;
    BumpMapID,SpecularMapID,DiffuseMapID : Cardinal;

    constructor Create(AOwner : TCollection); override; // for inital values
end;

Type PShader = ^TShader;

Type
NArray = Array Of String;

Type TShaderSrc = record
  Name,
  Data,
  Shader : String;
end;

TShaderError = record
ShaderID : Cardinal;
Layer : Cardinal; // 0 = no layer
ShaderName : String;
ErrorMessage : String;
end;

TShaderErrors = record
ShaderError : Array of TShaderError;
ShaderError_No : Integer;
end;

TShaderManager = class(TCollection)
  private
    FGeneralTexture : cardinal;
    FAvailableShaders : TStringList;
    FRequiredShaders : TStringList;
    FShadersSurfaceParams : TStringList;
    FLoadedTextures : TIntegerHash;
    FPath : string;
    function GetItem(Index : integer) : TShader;
  {  function ScanShader(path, filename : string) : boolean; // get available shaders
    }Function ScanShaderDir(Const dir,fullpath : String; Var Names : NArray; Var Name_No : Integer) : Boolean;
    Function ScanShaders(Const Base : String; Var Names : NArray; Var Name_No : Integer) : Boolean;
    {function ReadShadersFromList(filename : string; list : TStringList) : boolean;
    }
    function ReadShader(filename : string) : boolean;
    procedure ReadAvalibleShaders;
    function ScanAvailableShaders : boolean;
  public
    ShaderSrc : array of TShaderSrc; // Stores all the shaders source
    ShaderSrc_No : Integer;
    ShaderErrors : TShaderErrors; // Records Errors when loading shaders
    constructor Create(quakePath : string);
    destructor Destroy; override;
    Procedure AddRequiredShader(TextureName : String);
    Procedure ReadRequiredShaders;
    Function FindShaderSrc(TextureName : String) : Integer;
    function ReadShadersFromString(name,Data : string) : boolean;
    Procedure ProcessShaderTextures;
    function IndexOf(name : string) : integer;
    procedure SetVertexCoords_Shader(Const shaderId, numOfVerts : integer; var Data : array of TBSPVertex);
    procedure SetVertexColors_Shader(shaderId, stageId, numOfVerts : integer; var Data : array of TBSPVertex);
    procedure SetTextureCoordinates_Shader(Const shaderId, stageId, numOfVerts : integer; var Data : array of TBSPVertex);
    {function ReadRequiredShaders : boolean;
    function ReadShader(ShaderData, ShaderName : String) : boolean;
    }function LoadTexture(name : string;  NoPicMip, NoMipMap : boolean) : GLUINT;

    Procedure DrawFaces(var Vertexs : array of TBSPVertex; Const CenterOfModel, Position : TVector3f; Vertex_No : Integer; ShaderId : Integer; UseNormals : Boolean; sTan,tTan : Array of TVector3f);
    Procedure AddToShaderManager(Texture : string);
    function GetTextureNameInShaderFormat(Texture : String) : string;
    Function GetShaderId(Filename : String) : Integer;

    Procedure ClearShaderErrors;
    Procedure AddShaderError(ShaderID : Cardinal; ShaderName, ErrorMessage : String; Layer : Cardinal);

    Procedure UnloadTextures;
    procedure SetState(ShaderID, StageID : integer);
    Procedure CleanShader(var Shader : TStringList);

    property Items[Index: Integer]: TShader read GetItem; default;
    property AvailableShaders : TStringList read FAvailableShaders;
    property RequiredShaders : TStringList read FRequiredShaders write FRequiredShaders;
    property LoadedTextures : TIntegerHash read FLoadedTextures write FLoadedTextures;
end;

function Eval(wave : TWaveFunc) : single;
Function GetD3BumpMapFromAddNormals(S : String) : String;

implementation

uses ShaderValues,Textures,StrUtils,Helios_Timer{$IfDef HELIOS_APP},
     Helios_Global,Helios_Filesystem,Q3BSP, Helios_Roq_Engine{$Else},Helios_Basic{$EndIf},Helios_VertexFragmentProgram,Math;

procedure ChangeGamma(var color : array of byte; factor : single);
var i, j : integer;
    scale, r, g, b : single;
begin
  scale := 1;
  // extract the current RGB values
  r := color[0];
  g := color[1];
  b := color[2];

  // Multiply the factor by the RGB values, while keeping it to a 255 ratio
  r := r * factor / 255.0;
  g := g * factor / 255.0;
  b := b * factor / 255.0;

  // Check if the the values went past the highest value
  if (r > 1.0) and ((1.0/r) < scale) then scale:=(1.0/r);
  if (g > 1.0) and ((1.0/g) < scale) then scale:=(1.0/g);
  if (b > 1.0) and ((1.0/b) < scale) then scale:=(1.0/b);

  // Get the scale for this pixel and multiply it by our pixel values
  scale:=scale*255.0;
  r:=r*scale;	g:=g*scale;	b:=b*scale;

  // Assign the new gamma'nized RGB values to our image
  color[0] := Trunc(r);
  color[1] := Trunc(g);
  color[2] := Trunc(b);
end;

function removeExt(const s : string) : string;
var p : integer;
begin
  p := Pos('.tga', s);
  if p > 0 then begin
    result := Copy(s, 1, p-1);
    exit;
  end;

  p := Pos('.jpg', s);
  if p > 0 then begin
    result := Copy(s, 1, p-1);
    exit;
  end
  else
    result := s;
end;

function RemoveComment(txt : string) : string;
var p : integer;
    tmp : string;
begin
  txt := Trim(Txt);
  txt := StringReplace(txt, #9,  ' ', [rfReplaceAll]);

  tmp := txt;
  p := Pos('//', txt);
  if p > 0 then
    tmp := Trim(Copy(txt, 1, p-1));
  p := Pos('#', txt);
  if p > 0 then
    tmp := Trim(Copy(txt, 1, p-1));
  result := Trim(tmp);
end;

function GetToken(var txt : string) : string;
var l, p : integer;
    tmp : string;
begin
  l := Length(txt);
  tmp := '';

  p := Pos(' ', txt);
  if p > 0 then begin  // found
    result := Trim(Copy(txt, 1, p-1));
    txt := Trim(Copy(txt, p+1, l-p));
  end
  else begin // not found
    result := Trim(txt);
    txt := '';
  end;
end;

procedure Tokenize(txt : string; Tokens : TStringList);
var tok : string;
begin
  Tokens.Clear;
  repeat
    tok := GetToken(txt);
    if tok <> '' then
      Tokens.Add(tok);
  until txt = '';
end;

function IsNumeric(s : string) : boolean;
var i, l : integer;
begin
  if Length(s) > 0 then begin
    result := true;
    for i := 1 to Length(s) do begin
      if (MidStr(s, i, 1) < '0') or (MidStr(s, i, 1) > '9') then begin
        result := false;
        exit;
      end;
    end;
  end
  else
    result := false;
end;

function ConvertPath(p : string) : string;
begin
  result := StringReplace(p, '/', '\', [rfReplaceAll]);
end;

function InArray(Strings : Array Of String; Num : Integer; Value : String) : boolean;
var
x : integer;
begin
Result := False;

for x := 0 to Num-1 do
if ansilowercase(value) = ansilowercase(Strings[x]) then
Result := True;

end;

function SplitList(var txt : string) : string;
var l, p : integer;
    tmp : string;
begin
  l := Length(txt);
  tmp := '';

  p := Pos('=', txt);
  if p > 0 then begin  // found
    result := Trim(Copy(txt, 1, p-1));
    txt := Trim(Copy(txt, p+1, l-p));
  end
  else begin // not found
    result := Trim(txt);
    txt := '';
  end;
end;

function Eval(wave : TWaveFunc) : single;
var x, y : single;
begin
  // Evaluate a number of time based periodic functions
  // y = p[1] + p[2] * func( (time + p[3]) * p[4] )
  // y =  base + amp * func(timer + phase*freq)

  x := (gTimer.TotalTime  + wave.phase) * wave.freq;
//  x := gTimer.TotalTime  + wave.phase * wave.freq;
  x := x - Trunc(x);		// normalized

  case wave.func of
    wf_Sin :
      y := Sin(x * M_TWO_PI);
    wf_Triangle :
      if (x < 0.5) then
        y := 4.0 * x - 1.0
      else
        y := -4.0 * x + 3.0;
    wf_Square :
	if (x < 0.5) then
          y := 1.0
	else
          y := -1.0;
    wf_SawTooth :
      y := x;
    wf_InvSawTooth :
      y := 1.0 - x;
    wf_Noise :
      y := 1.0; // NOP
  end;
  result :=  y * wave.amp + wave.base;
end;

function RemoveEnd(S : String) : String;
var
x,y : integer;
begin

y := length(s);

for x := 0 to length(S)-1 do
If s[length(s)-x] <> ' ' then
begin
y := length(s)-x;
break;
end;

Result := Copy(S,1,y);

end;

// TShader *********************************************************************
constructor TShader.Create(AOwner : TCollection);
begin
  inherited Create(AOwner);
end;

// TShaderManager *********************************************************************
constructor TShaderManager.Create(quakePath : string);
begin
  inherited Create(TShader);

  FAvailableShaders := TStringList.Create;
  FRequiredShaders := TStringList.Create;
  FRequiredShaders.Sorted := true;
  FLoadedTextures := TIntegerHash.Create;

  FPath := IncludeTrailingPathDelimiter(quakePath);

  ScanAvailableShaders;
  ReadAvalibleShaders;
end;

Procedure UnloadTexture(ID : cardinal);
begin
If (Id <= 0) then exit;
glDeleteTextures(1,@ID);

end;

Procedure TShaderManager.UnloadTextures;
var
x,y,z :integer;
begin

If Count < 1 then exit;
exit;

for x := 0 to Count-1 do
begin
UnloadTexture(items[x].qerId);
UnloadTexture(items[x].BumpMapID);
UnloadTexture(items[x].DiffuseMapID);
UnloadTexture(items[x].SpecularMapID);

for y := 0 to items[x].numOfLayers-1 do
for z := 0 to 7 do
UnloadTexture(items[x].Layers[y].TextureId[z]);
end;

end;

destructor TShaderManager.Destroy;
begin
  UnloadTextures;
  FAvailableShaders.Free;
  FRequiredShaders.Free;
  FLoadedTextures.Free;

  inherited Destroy;
end;

function TShaderManager.GetItem(Index : integer) : TShader;
begin
  result := inherited Items[Index] as TShader;
end;

Function GetParam(Params : TStringList; ID : Integer) : String;
begin

Result := '0';

If ID < 0 then exit;
If ID > Params.Count-1 then exit;

Result := Params[ID];

end;

function TShaderManager.ReadShadersFromString(name,Data : string) : boolean;
var f : Text;
    line, tok, remain : string;
    stage : integer;
    curLayer : integer;
    newItem : TShader;
    tmpInt, i : integer;
    params,DataList : TStringList;
    required : boolean;
    src : string;
    modIdx, rgbIdx : integer;
    cast1, cast2 : integer;
    CLine : Integer;
label splitLine;
begin
  result := false;
  params := TStringList.Create;
  DataList := TStringList.Create;
  DataList.Text := Data;
  stage := 0;
  src := '';


      begin
        src := '';
        curLayer := 0;
        newItem := TShader(Add);
        newItem.Name := name;
        newItem.ShaderFile := 'NONE';
        newItem.numOfLayers := 0;
        newItem.NoMipMap := false;
        newItem.NoPicMip := false;
        newItem.Cull := GL_FRONT;
        newItem.qerId := 0;
        newItem.qerImage := '';
        newItem.HasFog := false;
        newItem.IsSkyBox := false;
        newItem.TotallyTrans := false;
        newItem.SortKey := sk_opaque;
        newItem.effectIndex := -1;
        newItem.SkyboxType := -1;
        newItem.polygonOffset := False;
        newItem.HSurface := 0;
      end;

  Stage := 1;
  CLine := 0;
  required := true;

  while CLine < DataList.Count do begin
  if CLine >= DataList.Count then break;
    Line := DataList.Strings[CLine];
    Inc(CLine);

    src := src + line + #13#10;

splitLine:

    line := RemoveComment(line);

    if (Copy(line, 1, 2) = '//') then continue;

    if Copy(line, 1, 1) = '{' then begin
      inc(stage);
      if (stage = 2) and required then begin
        inc(curLayer);
        SetLength(newItem.Layers, curLayer);
        // initial values
        newItem.Layers[curLayer-1].DepthFunc := GL_LEQUAL; //Lookup('lequal');
        newItem.Layers[curLayer-1].DepthWrite := true;
        newItem.Layers[curLayer-1].AlphaFunc := Lookup('noalpha');
        newItem.Layers[curLayer-1].AlphaRef := 1;
        newItem.Layers[curLayer-1].AlphaWave.func := wf_None;
        newItem.Layers[curLayer-1].doAlphaTest := false;
        newItem.Layers[curLayer-1].doBlending := false;
        newItem.Layers[curLayer-1].SrcBlend := GL_SRC_ALPHA;
        newItem.Layers[curLayer-1].DestBlend := GL_ONE_MINUS_SRC_ALPHA;
        newItem.Layers[curLayer-1].TexClamp := false;
        newItem.Layers[curLayer-1].numOfrgbGen := 0;
        newItem.Layers[curLayer-1].numOftcMods := 0;
        newItem.Layers[curLayer-1].UseVertexColors := false;
      end;
      if Length(line) > 1 then begin
        line := Trim(Copy(line, 2, Length(line)-1));
        goto splitLine;
      end
      else
        continue;
    end;
    if Copy(line, 1, 1) = '}' then begin
      dec(stage);
      if (stage = 1) and required then begin
        newItem.numOfLayers := curLayer;
      end;

      // copy last shader source
      if (stage = 0) and required then begin
        newItem.Source := src;
        src := '';
      end;

      if (stage = 0) then
        src := ''; // reset tmp source
      continue;
    end;

    if (Length(line) > 0) and required then begin
      line := LowerCase(line);
      Tokenize(line, params);

      if params[0] = 'sort' then
      if params[1] = 'decal' then
      newItem.polygonOffset := True;
      
      if params[0] = 'polygonoffset' then
      newItem.polygonOffset := True;

    if (stage = 1) then begin
      if params[0] = 'surfaceparm' then begin
          try
          {If params[1] = 'sky' then
          begin
          newItem.IsSkyBox := true;
          newItem.SkyboxType := 1;
          continue;
          end;}

          //FShadersSurfaceParams.Add(params[1]);
          newItem.HSurface := {newItem.HSurface + }LookUpHeliosSurface(params[1]);

          if params[1] = 'nonopaque' then
          newItem.NonOpaque := true
          else
          begin

          newItem.Surface := newItem.Surface + LookUp(params[1]);

          if newItem.Surface < 0 then
          newItem.Surface := 0;
          end;
          except
            //ShowMessage(params[1]);
          end;
          continue;
        end;

        if params[0] = 'cull' then begin
           newItem.Cull := LookUp(params[1]);
           continue;
        end;

        if params[0] = 'sort' then begin
          tmpInt := LookUp(params[1]);
          if tmpInt >   -1 then
            newItem.SortKey := enSortKey(tmpInt);
          continue;
        end;

        if params[0] = 'skyparms' then begin
          // set the global skyBox class
          newItem.IsSkyBox := true;
          skyBoxId := Count-1;

          if (params[1] = '-') or IsNumeric(params[1]) or (params[1] = 'full') then begin // type = sphere
            newItem.SkyboxType := 1;
          end
          else begin                   // type = box
            // found in q3dm10, q3tourney1
            newItem.SkyboxType := 2;
            SetLength(newItem.SkyboxTextureNames, 6);
            SetLength(newItem.SkyboxTextureIds, 6);
            newItem.SkyboxTextureNames[0] := ConvertPath(params[1]+'_rt');
            newItem.SkyboxTextureNames[1] := ConvertPath(params[1]+'_lf');
            newItem.SkyboxTextureNames[2] := ConvertPath(params[1]+'_ft');
            newItem.SkyboxTextureNames[3] := ConvertPath(params[1]+'_bk');
            newItem.SkyboxTextureNames[4] := ConvertPath(params[1]+'_up');
            newItem.SkyboxTextureNames[5] := ConvertPath(params[1]+'_dn');
            for i := 0 to 5 do
              newItem.SkyboxTextureIds[i] := 0;
          end;
          continue;
        end;

        if params[0] = 'deformvertexes' then begin
          inc(newItem.numOfDeformVertexes);
          SetLength(newItem.DeformVertexes, newItem.numOfDeformVertexes);
          newItem.DeformVertexes[newItem.numOfDeformVertexes-1].deformType := enDeformType(Lookup('def_'+params[1]));
          with newItem.DeformVertexes[newItem.numOfDeformVertexes-1] do begin
            case deformType of
              DEFORMVERTEXES_AUTOSPRITE,
              DEFORMVERTEXES_AUTOSPRITE2 :
                newItem.Cull := GL_NONE;
              DEFORMVERTEXES_WAVE :
                begin
                  for i := 0 to min(5,max(params.Count-3,0)) do
                    if i = 1 then
                      Values[i] := Lookup(params[2+i])
                    else
                      Values[i] := StrToFloat(params[2+i]);
                end;
              DEFORMVERTEXES_NORMAL :
                begin

                // Warning some ctf shaders only have 2 parameters!!!

                  // wave func at i=1
                  for i := 0 to min(4,max(params.Count-3,0)) do
                    if i = 1 then
                      Values[i] := Lookup(params[2+i])
                    else
                      Values[i] := StrToFloat(params[2+i]);
                end;
              DEFORMVERTEXES_BULGE :
                begin
                  for i := 0 to min(2,max(params.Count-3,0)) do
                    Values[i] := StrToFloat(params[2+i]);
                end;
              DEFORMVERTEXES_MOVE :
                begin
                  for i := 0 to min(7,max(params.Count-3,0)) do
                    if i = 3 then
                      Values[i] := Lookup(params[2+i])
                    else
                      Values[i] := StrToFloat(params[2+i]);
                end;
            end;
          end;
          continue;
        end;

        if (params[0] = 'fogparms') or (params[0] = 'fogvars') then begin
          newItem.HasFog := true;
          // color params with parens !!!
          newItem.Fog.Red := StrToFloat(GetParam(Params,2));
          newItem.Fog.Green := StrToFloat(GetParam(Params,3));
          newItem.Fog.Blue := StrToFloat(GetParam(Params,4));
          newItem.Fog.Alpha := 1;
          newItem.Fog.Distance := StrToFloat(GetParam(Params,6));

          continue;
        end;

       // If newItem.IsSkyBox then
        if params[0] = 'fogvars' then begin
        {  Fog.HasFog := True;
          Fog.FogColour[0] := StrToFloat(params[2]);
          Fog.FogColour[1] := StrToFloat(params[3]);
          Fog.FogColour[2] := StrToFloat(params[4]);
          Fog.FogColour[3] := 1;
          Fog.FogMin := StrToFloat(params[6]);
          continue; }
        end;

        if params[0] = 'qer_editorimage' then begin
          newItem.qerImage := ConvertPath(params[1]);
          continue;
        end;

        if params[0] = 'bumpmap' then begin
          newItem.BumpMap := ConvertPath(GetD3BumpMapFromAddNormals(params[1]));
          continue;
        end;

        if params[0] = 'specularmap' then begin
          newItem.SpecularMap := ConvertPath(params[1]);
          continue;
        end;

        if params[0] = 'diffusemap' then begin
          newItem.DiffuseMap := ConvertPath(params[1]);
          continue;
        end;

        if params[0] = 'nopicmip' then begin
          newItem.NoPicMip := true;
          continue;
        end;
        if params[0] = 'nomipmap' then begin
          newItem.NoMipMap := true;
          continue;
        end;

        if params[0] = 'portal' then begin
          continue;
        end;
      end // (stage = 1)
      else if (Stage = 2) then begin
        if params[0] = 'depthfunc' then begin
          newItem.Layers[curLayer-1].DepthFunc := LookUp(params[1]);
          if newItem.Layers[curLayer-1].DepthFunc = -1 then
            newItem.Layers[curLayer-1].DepthFunc := GL_LEQUAL;
          continue;
        end;
        if params[0] = 'depthwrite' then begin
          if (curLayer = 1) and (newItem.TotallyTrans) then
            newItem.TotallyTrans := false;
          newItem.Layers[curLayer-1].DepthWrite := true;
          continue;
        end;
        if params[0] = 'map' then begin
          newItem.Layers[curLayer-1].animFreq := 0; // no animation
          newItem.Layers[curLayer-1].animFrameTime := 0;
          newItem.Layers[curLayer-1].numOfTextures := 1;
          SetLength(newItem.Layers[curLayer-1].TextureNames,1);
          If params[1] = 'clamp' then
          begin
          newItem.Layers[curLayer-1].TextureNames[0] := ConvertPath(params[2]);
          newItem.Layers[curLayer-1].TexClamp := true;
          end
          else
          newItem.Layers[curLayer-1].TextureNames[0] := ConvertPath(params[1]);
          continue;
        end;
        if params[0] = 'clampmap' then begin
          newItem.Layers[curLayer-1].animFreq := 0; // no animation
          newItem.Layers[curLayer-1].animFrameTime := 0;
          newItem.Layers[curLayer-1].numOfTextures := 1;
          SetLength(newItem.Layers[curLayer-1].TextureNames,1);
          newItem.Layers[curLayer-1].TextureNames[0] := ConvertPath(params[1]);
          newItem.Layers[curLayer-1].TexClamp := true;
          continue;
        end;
        if params[0] = 'animmap' then begin
          newItem.Layers[curLayer-1].animFreq := StrToFloat(params[1]); // animation speed
          newItem.Layers[curLayer-1].animFrameTime := Round(1000.0 / newItem.Layers[curLayer-1].animFreq);

          newItem.Layers[curLayer-1].numOfTextures := params.Count-2;
          SetLength(newItem.Layers[curLayer-1].TextureNames,params.Count-2);
          for i := 0 to params.Count-3 do begin
            newItem.Layers[curLayer-1].TextureNames[i] := ConvertPath(params[2+i]);
          end;
          continue;
        end;
        if params[0] = 'detail' then begin
          continue;
        end;
        if params[0] = 'alphafunc' then begin
          newItem.Layers[curLayer-1].AlphaFunc := Lookup(params[1]);
          newItem.Layers[curLayer-1].doAlphaTest := true;
          newItem.Layers[curLayer-1].DepthWrite := false;
          if newItem.Layers[curLayer-1].AlphaFunc = GL_GREATER then
            newItem.Layers[curLayer-1].AlphaRef := 0
          else
            newItem.Layers[curLayer-1].AlphaRef := 0.5;
          continue;
        end;
        if params[0] = 'alphagen' then begin
          if params.Count = 5 then begin // must be a wave
            with newItem.Layers[curLayer-1].AlphaWave  do begin
              func := enWave(Lookup(GetParam(Params,2)));
              base := StrToFloat(GetParam(Params,3));
              amp := StrToFloat(GetParam(Params,4));
              phase := StrToFloat(GetParam(Params,5));
              freq := StrToFloat(GetParam(Params,6));
            end;
          end;
          continue;
        end;
        
        if params[0] = 'rotate' then begin
        inc(newItem.Layers[curLayer-1].numOftcmods);
        SetLength(newItem.Layers[curLayer-1].tcmods, newItem.Layers[curLayer-1].numOftcmods);
        modIdx := newItem.Layers[curLayer-1].numOftcmods-1;

        newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].modType := entcmod(Lookup('rotate'));
        newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[0] := StrToFloat(params[1]); // rotate
        end;

        if params[0] = 'scale' then begin
        inc(newItem.Layers[curLayer-1].numOftcmods);
        SetLength(newItem.Layers[curLayer-1].tcmods, newItem.Layers[curLayer-1].numOftcmods);
        modIdx := newItem.Layers[curLayer-1].numOftcmods-1;

        newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].modType := entcmod(Lookup('scale'));

         for i := 0 to 1 do
          If StrToFloat(params[1+i]) > 1 then
          newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[i] := 1/StrToFloat(params[1+i])
          else
          newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[i] := StrToFloat(params[1+i]);
        end;

        if params[0] = 'tcmod' then begin
          inc(newItem.Layers[curLayer-1].numOftcmods);
          SetLength(newItem.Layers[curLayer-1].tcmods, newItem.Layers[curLayer-1].numOftcmods);
          modIdx := newItem.Layers[curLayer-1].numOftcmods-1;
          tmpInt := Lookup(params[1]);
          if tmpInt = -1 then
            tmpInt := 0;
          newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].modType := entcmod(tmpInt);
          case tmpInt of
            1 : newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[0] := StrToFloat(GetParam(Params,2)); // rotate
            2 : begin // scale
                    for i := 0 to 1 do
                    If StrToFloat(params[2+i]) > 1 then
                      newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[i] := 1/StrToFloat(GetParam(Params,2+i))
                      else
                      newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[i] := StrToFloat(GetParam(Params,2+i));
                  end;
            3 : begin // scroll
                    for i := 0 to 1 do
                      newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[i] := StrToFloat(GetParam(Params,2+i));
                  end;
            4 : begin // stretch
                  with newItem.Layers[curLayer-1].tcMods[modIdx].Wave  do begin
                    func := enWave(Lookup(GetParam(Params,2)));
                    base := StrToFloat(GetParam(Params,3));
                    amp := StrToFloat(GetParam(Params,4));
                    phase := StrToFloat(GetParam(Params,5));
                    freq := StrToFloat(GetParam(Params,6));
                  end;
                end;
            5 : begin // transform
                  for i := 0 to 5 do
                    newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[i] := StrToFloat(params[2+i]);
                end;
            6 : begin // turb
                  // first param is unused
                  newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[0] := 0; //StrToFloat(params[2]);
                  newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[1] := StrToFloat(GetParam(Params,3));
                  newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[2] := StrToFloat(GetParam(Params,4));
                  newItem.Layers[curLayer-1].tcmods[newItem.Layers[curLayer-1].numOftcmods-1].Values[3] := StrToFloat(GetParam(Params,5));
                end;
          end;
          continue;
        end;
        if params[0] = 'rgbgen' then begin
          rgbIdx := newItem.Layers[curLayer-1].numOfrgbgen;
          inc(newItem.Layers[curLayer-1].numOfrgbgen);
          SetLength(newItem.Layers[curLayer-1].rgbGen, rgbIdx+1);
          newItem.Layers[curLayer-1].rgbGen[rgbIdx].rgbGType := enRGBGen(LookUp(params[1]));

          case newItem.Layers[curLayer-1].rgbGen[rgbIdx].rgbGType of
            rgbGen_Wave :
              begin
                with newItem.Layers[curLayer-1].rgbGen[rgbIdx].Wave  do begin
                  func := enWave(Lookup(GetParam(Params,2)));
                  base := StrToFloat(GetParam(Params,3));
                  amp := StrToFloat(GetParam(Params,4));
                  phase := StrToFloat(GetParam(Params,5));
                  freq := StrToFloat(GetParam(Params,6));
                end;
                newItem.Layers[curLayer-1].UseVertexColors := false;
                FillChar(newItem.Layers[curLayer-1].curColor, 4, 255); // white
              end;
            rgbGen_Identity :
              begin
                newItem.Layers[curLayer-1].UseVertexColors := false;
                FillChar(newItem.Layers[curLayer-1].curColor, 4, 255);
              end;
            rgbGen_IdentityLightning :
              begin
                newItem.Layers[curLayer-1].UseVertexColors := false;
                FillChar(newItem.Layers[curLayer-1].curColor, 4, 128);
                ChangeGamma(newItem.Layers[curLayer-1].curColor, {$IfDef HELIOS_APP}HeliosSettings.gamma{$Else}StaticGamma{$EndIf});

              end;
            rgbGen_Entity,
            rgbGen_OneMinusEntity : // @@@@ must be adjusted if Entity data is available
              begin
                newItem.Layers[curLayer-1].UseVertexColors := false;
                FillChar(newItem.Layers[curLayer-1].curColor, 4, 255); // white
              end;
           rgbGen_Vertex,
           rgbGen_OneMinusVertex :  newItem.Layers[curLayer-1].UseVertexColors := true;
           rgbGen_LightningDiffuse : newItem.Layers[curLayer-1].UseVertexColors := true; // @@@ eval in setColor
          end;
          continue;
        end;

        if params[0] = 'tcgen' then begin
          tmpInt := LookUp(params[1]);
          if tmpInt = -1 then begin
            newItem.Layers[curLayer-1].tcGen := tcGen_standard;
            // param - array is missing for this function
          end
          else
            newItem.Layers[curLayer-1].tcGen := entcGen(tmpInt);
          continue;
        end;

        if params[0] = 'hsurface' then begin
          if ansilowercase(params[1]) = 'dynamic_texture' then
          newItem.Layers[curLayer-1].DynamicTexture := True;
          continue;
        end;

        if params[0] = 'videomap' then
        begin
          newItem.Layers[curLayer-1].DynamicTexture := True;
          newItem.Layers[curLayer-1].animFreq := 0; // no animation
          newItem.Layers[curLayer-1].animFrameTime := 0;
          newItem.Layers[curLayer-1].numOfTextures := 1;
          SetLength(newItem.Layers[curLayer-1].TextureNames,1);
          newItem.Layers[curLayer-1].TextureNames[0] := ConvertPath(params[1]);
          continue;

        end;

        if (params[0] = 'blendfunc') or (params[0] = 'blend') then begin
          newItem.Layers[curLayer-1].doBlending := true;
          If (params[1] = 'bumpmap') then
          begin
          newItem.Layers[curLayer-1].IsBumpMap := True;
          newItem.Layers[curLayer-1].doBlending := False;
          end
          else
          If (params[1] = 'diffusemap') then
          begin
          newItem.Layers[curLayer-1].IsDiffuseMap := True;
          newItem.Layers[curLayer-1].doBlending := False;
          end
          else
          if (params[1] = 'add') or (params[1] = 'gl_add')then begin
            newItem.Layers[curLayer-1].SrcBlend := GL_ONE;
            newItem.Layers[curLayer-1].DestBlend := GL_ONE;
            newItem.Layers[curLayer-1].DepthWrite := false;
          end
          else if params[1] = 'filter' then begin
            newItem.Layers[curLayer-1].SrcBlend := GL_DST_COLOR;
            newItem.Layers[curLayer-1].DestBlend := GL_ZERO;
            newItem.Layers[curLayer-1].DepthWrite := false;
          end
          else if params[1] = 'blend' then begin
            newItem.Layers[curLayer-1].SrcBlend := GL_SRC_ALPHA;
            newItem.Layers[curLayer-1].DestBlend := GL_ONE_MINUS_SRC_ALPHA;
            newItem.Layers[curLayer-1].DepthWrite := false;
          end
          else
          If params.Count > 2 then
          begin
            newItem.Layers[curLayer-1].SrcBlend :=  LookUp(params[1]);
            newItem.Layers[curLayer-1].DestBlend := LookUp(params[2]);
            newItem.Layers[curLayer-1].DepthWrite := false;
{            newItem.Layers[curLayer-1].DepthWrite := (newItem.Layers[curLayer-1].SrcBlend = GL_ONE) and
                                                     (newItem.Layers[curLayer-1].SrcBlend = GL_ZERO) and
                                                     (curLayer = 1);}

          end;
          If newItem.Layers[curLayer-1].doBlending then
          if (curLayer = 1)  then begin
            if (newItem.Layers[curLayer-1].DestBlend = GL_ONE) AND (newItem.Layers[curLayer-1].DestBlend = GL_ZERO) then
              newItem.Layers[curLayer-1].doBlending := false;
            newItem.TotallyTrans := newItem.Layers[curLayer-1].DestBlend <> GL_ZERO;
          end;
          continue;
        end; // end blendfunc
      end;
    end;
  end;
  params.Free;
  DataList.Free;
  result := true;
end;

Procedure TShaderManager.CleanShader(var Shader : TStringList);
var line : string;
    stage : integer;
    src : string;
    CLine : Integer;
begin

  stage := 0;
  src := '';
  CLine := 0;
  while CLine < Shader.Count do begin
  if CLine >= Shader.Count then break;
    Line := Shader.Strings[CLine];
    Inc(CLine);

    Line := RemoveComment(Line);

    // Ignore Doom3 Table References
    if (Pos('table ',ansilowercase(line)) > 0) then
    Line := '';

    if stage = 0 then // Doom 3 fix
    if (copy(line,1,1) <> ' ') and (copy(line,1,1) <> ' ') and (copy(line,1,1) <> '{') then
    if (Pos('{',line) > 0) then
    begin
    StringReplace(Line,'{','',[rfReplaceAll]);
    src := src + #13 + Line + #13 + '{';
    Line := '';
    inc(stage);
    end;

    if (Pos('{',line) > 0) then
      inc(stage);

    if (Pos('}',line) > 0) then
      dec(stage);

    if line = 'textures/REGION' then
    line := line;

    If ((Line = '') and (stage = 0)) or (line = '') then
    else
    begin
    If src > '' then
    src := src + #13+ Line
    else
    src := Line;
    end;
  end;
  shader.Text := Src;
end;

function TShaderManager.ReadShader(filename : string) : boolean;
var line : string;
    stage : integer;
    Shader : TStringList;
    required : boolean;
    src,Next,ShaderFile : string;
    CLine : Integer;
label splitLine;
begin
  result := false;
  Shader := TStringList.Create;
  Shader.LoadFromFile(filename);
  CleanShader(Shader);
  ShaderFile := ansilowercase(Copy(extractfilename(Filename),1,length(extractfilename(Filename))-length(extractfileext(Filename))));
  stage := 0;
  src := '';
  {ShaderSrc_No := 0;
  SetLength(ShaderSrc,ShaderSrc_No);}
  CLine := 0;
  while CLine < Shader.Count do begin
  if CLine >= Shader.Count then break;
    Line := Shader.Strings[CLine];
    Inc(CLine);
  //  src := src + line + #13#10;

splitLine:

    line := RemoveComment(line);

    if (Copy(line, 1, 2) = '//') then continue;

    If CLine < Shader.Count-1 then
    Next := Shader.Strings[CLine]
    else
    Next := '';

    if (line <> '') and (stage=0) and (line <> '{') and (Pos('{',next) > 0) then begin // must be the shader name
      line := RemoveEnd(StringReplace(ConvertPath(line), '{',  '', [rfReplaceAll]));
      Inc(ShaderSrc_No);
      SetLength(ShaderSrc,ShaderSrc_No);
      ShaderSrc[ShaderSrc_No-1].Name := Line;
      ShaderSrc[ShaderSrc_No-1].Shader := ShaderFile;
      continue;
    end;

    If ShaderSrc_No < 1 then continue;

    if (Pos('{',line) > 0) then begin
      inc(stage);
      If stage > 1 then
      src := src + #13+ Line;
        continue;
    end;

    if (Pos('}',line) > 0) then begin
      dec(stage);
      if (stage = 0) then
      begin
      ShaderSrc[ShaderSrc_No-1].Data := src;
      src := ''; // reset tmp source
      continue;
      end;
    end;

    If stage > 0 then
    begin
    If src > '' then
    src := src + #13+ Line
    else
    src := Line;
    end;
  end;
  shader.Free;
  result := true;
end;

Procedure TShaderManager.ReadAvalibleShaders;
var
x : integer;
TMP : String;
begin

//Clear the source!
ShaderSrc_No := 0;
SetLength(ShaderSrc,ShaderSrc_No);

for x := 0 to FAvailableShaders.Count-1 do
ReadShader(FAvailableShaders.Strings[x]);

end;

Function TShaderManager.ScanShaderDir(Const dir,fullpath : String; Var Names : NArray; Var Name_No : Integer) : Boolean;
var
sr: TSearchRec;
begin

Result := False;

    if FindFirst(dir, faAnyFile, sr) = 0 then begin
    repeat
    If not InArray(Names,Name_No,sr.Name) then
    Begin
    FAvailableShaders.Add(fullpath + sr.Name);
      inc(Name_No);
      SetLength(Names,Name_No);
      Names[Name_No-1] := sr.Name;
    end;
    until FindNext(sr) <> 0;
    FindClose(sr);
    result := true;
  end;

end;

Function TShaderManager.ScanShaders(Const Base : String; Var Names : NArray; Var Name_No : Integer) : Boolean;
Var
Fullpath,dir : string;
begin
  Result := False;

  fullpath := Base + '\scripts\';
  dir := Base + '\scripts\*.shader';
  If ScanShaderDir(dir,fullpath,Names,Name_No) then
  Result := True;

  fullpath := Base + '\shaders\';
  dir := Base + '\shaders\*.shader';
  If ScanShaderDir(dir,fullpath,Names,Name_No) then
  Result := True;

  // Doom 3
  fullpath := Base + '\materials\';
  dir := Base + '\materials\*.mtr';
  If ScanShaderDir(dir,fullpath,Names,Name_No) then
  Result := True;

end;

function TShaderManager.ScanAvailableShaders : boolean;
var sr: TSearchRec;
    fullpath, dir : string;
    Names : NArray;
    Name_No : integer;
begin
  result := false;
  if Count > 0 then
  begin
  UnloadTextures;
  Clear;
  end;

  FAvailableShaders.Clear;
  Name_No := 0;
  SetLength(Names,Name_No);

  If MOD_FOLDER <> '' then
  If ScanShaders(GAME_FOLDER + '\' + MOD_FOLDER,Names,Name_No) then
  Result := True;

  If BaseDir <> '' then
  If ScanShaders(GAME_FOLDER + '\' + Basedir,Names,Name_No) then
  Result := True;

  If ScanShaders(GAME_FOLDER + '\' + HELIOS_MODFOLDER,Names,Name_No) then
  Result := True;

  If ScanShaders(GAME_FOLDER + '\',Names,Name_No) then
  Result := True;

  SetLength(Names,0);
end;

Procedure TShaderManager.AddRequiredShader(TextureName : String);
var
x : integer;
Texture : String;
begin

Texture := ansilowercase(TextureName);

If FRequiredShaders.Count > 0 then
for x := 0 to FRequiredShaders.Count-1 do
If Texture = FRequiredShaders.Strings[x] then
exit; // Already in list!

FRequiredShaders.Add(Texture);
end;

Procedure TShaderManager.ProcessShaderTextures;
var
i,j,t,TL : Integer;
Name : String;
begin
ClearShaderErrors;

// bind textures
  if Count > 0 then begin // any shaders loaded?
    for i := 0 to Count-1 do begin
      if Items[i].qerImage <> '' then begin
        items[i].qerId := LoadTexture(items[i].qerImage, Items[i].NoPicMip, Items[i].NoMipMap);
      end;

    {  If (Items[i].numOfLayers < 1) and (Items[i].qerImage <> '') then
      begin
      Items[i].numOfLayers := 1;
      SetLength(Items[i].Layers,Items[i].numOfLayers);
      Items[i].Layers[0].numOfTextures := 1;
      SetLength(Items[i].Layers[0].TextureNames,Items[i].Layers[0].numOfTextures);
      Items[i].Layers[0].TextureNames[0] := Items[i].qerImage;
      end;
      }

         {
      If (Items[i].numOfLayers < 1) and (Items[i].BumpMap <> '') then
      begin
      Items[i].numOfLayers := 1;
      SetLength(Items[i].Layers,Items[i].numOfLayers);
      Items[i].Layers[0].numOfTextures := 1;
      SetLength(Items[i].Layers[0].TextureNames,Items[i].Layers[0].numOfTextures);
      Items[i].Layers[0].TextureNames[0] := Items[i].BumpMap;
      end;                               }

      If (Items[i].BumpMap <> '') then
      Items[i].BumpMapID := LoadTexture(Items[i].BumpMap, False, True);

      If (Items[i].DiffuseMap <> '') then
      Items[i].DiffuseMapID := LoadTexture(Items[i].DiffuseMap, False, True);
      If (Items[i].SpecularMap <> '') then
      Items[i].SpecularMapID := LoadTexture(Items[i].SpecularMap, False, True);

      If Items[i].BumpMapID < 1 then
      If (Items[i].numOfLayers < 1) and (Items[i].DiffuseMap <> '') then
      begin
      Items[i].numOfLayers := 1;
      SetLength(Items[i].Layers,Items[i].numOfLayers);
      Items[i].Layers[0].numOfTextures := 1;
      SetLength(Items[i].Layers[0].TextureNames,Items[i].Layers[0].numOfTextures);
      Items[i].Layers[0].TextureNames[0] := Items[i].DiffuseMap;
      end;

      if Items[i].numOfLayers > 0 then begin
        for j := 0 to Items[i].numOfLayers-1 do begin
          if Items[i].Layers[j].numOfTextures > 0 then begin
          Items[i].Layers[j].Draw := False;
         { If Items[i].Layers[j].DynamicTexture then
          begin

          Items[i].Layers[j].TextureId[0] := LoadRoq_M(Items[i].Layers[j].TextureNames[0],True,False,True,1);
          If Items[i].Layers[j].TextureId[0] <> -1 then
          Items[i].Layers[j].TextureId[0] := Cinematics.Cinematic[Items[i].Layers[j].TextureId[0]].Texture;
          end
          else       }
          If Items[i].Layers[j].IsBumpMap then
          begin
          Items[i].BumpMap   := GetD3BumpMapFromAddNormals(Items[i].Layers[j].TextureNames[0]);
          Items[i].BumpMapID := LoadTexture(Items[i].BumpMap, False, True);
          end
          else
          If Items[i].Layers[j].IsDiffuseMap then
          begin
          Items[i].DiffuseMap   := Items[i].Layers[j].TextureNames[0];
          Items[i].DiffuseMapID := LoadTexture(Items[i].DiffuseMap, False, True);
          end
          else
            for t := 0 to Items[i].Layers[j].numOfTextures-1 do begin
              if Copy(Items[i].Layers[j].TextureNames[t], 1, 1) <> '$' then begin // must be a texture name
                name := Items[i].Layers[j].TextureNames[t];
                Items[i].Layers[j].TextureId[t] := LoadTexture(name, Items[i].NoPicMip, Items[i].NoMipMap);
                if Items[i].Layers[j].TextureId[t] = 0 then
                if Items[i].numOfLayers = 1 then
                Items[i].Layers[j].TextureId[t] := Items[i].DiffuseMapID;

                if Items[i].Layers[j].TextureId[t] = 0 then
                AddShaderError(i,Items[i].Name,'Error Loading Texture: ' + Name,j+1)
                else
                inc(TL);

         //         Items[i].Layers[j].TextureId[t] := items[i].qerId;//Items[i].Layers[j].TextureId[t];
                FLoadedTextures[RemoveExt(name)] := Items[i].Layers[j].TextureId[t];
                Items[i].Layers[j].UseLightmap := false;

                if Items[i].Layers[j].TextureId[t] > 0 then
                Items[i].Layers[j].Draw := True;
              end
              else begin
                Items[i].Layers[j].UseLightmap := true;
                if Items[i].Layers[j].TextureNames[t] = '$lightmap' then
                  FillChar(Items[i].Layers[j].curColor, 4, 255)
                else// if Items[i].Layers[j].TextureNames[t] = '$identity' then
                  FillChar(Items[i].Layers[j].curColor, 4, 128);
                if Items[i].Layers[j].TextureId[t] > 0 then
                Items[i].Layers[j].Draw := True;
              end;
              // if no rgbGen is available then add $identity
              if Items[i].Layers[j].numOfrgbGen = 0 then
                FillChar(Items[i].Layers[j].curColor, 4, 255);
            end;
          end;
        end;
      end;
      If TL = 0 then
    AddShaderError(i,Items[i].Name,'Error No Textures Loaded',0);
    end;
  end;

end;

Function TShaderManager.FindShaderSrc(TextureName : String) : Integer;
var
x : integer;
Texture : String;
begin
Result := -1;

Texture := GetTextureNameInShaderFormat(ansilowercase(TextureName));

for x := 0 to ShaderSrc_No-1 do
If ansilowercase(ShaderSrc[x].Name) = texture then
begin
Result := x;
Exit;
end;

end;

Procedure TShaderManager.ReadRequiredShaders;
var
x,ReqSS : integer;
begin

UnloadTextures;
Clear;
If ShaderSrc_No < 1 then exit;
If FRequiredShaders.Count < 1 then exit;

for x := 0 to FRequiredShaders.Count-1 do
begin
ReqSS := FindShaderSrc(FRequiredShaders.Strings[x]);
If ReqSS > -1 then
begin
ReadShadersFromString(ShaderSrc[ReqSS].Name,ShaderSrc[ReqSS].Data);
end;
end; // For End

If (Count > 0) then
ProcessShaderTextures;

end;

function TShaderManager.LoadTexture(name : string;  NoPicMip, NoMipMap : boolean) : GLUINT;
var noExt : string;
    newId : GLUINT;
    p : integer;
    name_only,FN : string;
    l : integer;
begin
  name := ansilowercase(StringReplace(name, '/', '\', [rfReplaceAll]));

  p := Pos('.tga', Name);
  if p > 0 then
    Name := Copy(Name, 1, p-1);
  p := Pos('.jpg', Name);
  if p > 0 then
    Name := Copy(Name, 1, p-1);
  p := Pos('.dds', Name);
  if p > 0 then
    Name := Copy(Name, 1, p-1);
  p := Pos('.bmp', Name);
  if p > 0 then
    Name := Copy(Name, 1, p-1);

  FN := DoesFileexist(Name+'.jpg');
  If FN = '' then
  FN := DoesFileexist(Name+'.tga');
  If FN = '' then
  FN := DoesFileexist(Name+'.dds');
  If FN = '' then
  FN := DoesFileexist(Name+'.bmp');
  If FN = '' then
  begin
  {$IfDef HELIOS_CONSOLE}If name <> '' then Cons.AddMsg('Couldn`t Load: ' + Name);{$EndIf}
  Result := 0;
  exit;
  end;

  Name := FN;

  noExt := Name;

  p := Pos('.tga', noExt);
  if p > 0 then
    noExt := Copy(noExt, 1, p-1);
  p := Pos('.jpg', noExt);
  if p > 0 then
    noExt := Copy(noExt, 1, p-1);
  p := Pos('.dds', noExt);
  if p > 0 then
    noExt := Copy(noExt, 1, p-1);
  p := Pos('.bmp', noExt);
  if p > 0 then
    noExt := Copy(noExt, 1, p-1);

  name_only := ExtractFileName(noExt);
  l := Length(name_only);
//  repeat
    if FLoadedTextures.Exists(noExt) then begin
      result := FLoadedTextures[noExt];
      exit;
    end ;

    // tex not loaded
    LoadQuakeTexture('', noExt, newId, NoPicMip, NoMipMap  );
    if newId <> 0 then begin
      FLoadedTextures[noext] := newId;
      result := newId;
      exit;
    end;
{    dec(l);
    noExt := Copy(noExt, 1, Length(noExt)-1);
  until l < 2;           }
  result := 0; // no texture
end;

function TShaderManager.IndexOf(name : string) : integer;
var i : integer;
begin
  result := -1;
  name := lowercase(name);
  if Count > 0 then begin
    for i := 0 to Count-1 do begin
      if LowerCase(Items[i].Name) = lowercase(name) then begin
        result := i;
        exit;
      end;
    end;
  end;
end;

procedure TShaderManager.SetVertexCoords_Shader(Const shaderId, numOfVerts : integer; var Data : array of TBSPVertex);
var shader : TShader;
    i, j, nv : integer;
    args : TWaveFunc;
    startoff, off, wavesize, inc : single;
    p, v, middle, d, h : TVector3f;
    st : TVector2f;
const speed = 2.0;
begin

  shader := Items[shaderId];
  if shader.numOfDeformVertexes > 0 then begin
    startoff := 0;
    wavesize := 1;
    for i := 0 to shader.numOfDeformVertexes-1 do begin
      case shader.DeformVertexes[i].deformType of
        DEFORMVERTEXES_WAVE :
          begin
            args.func := enWave(trunc(shader.DeformVertexes[i].values[1]));
            args.base := shader.DeformVertexes[i].values[2];
            args.amp  := shader.DeformVertexes[i].values[3];
            wavesize := shader.DeformVertexes[i].values[0];
            startoff := shader.DeformVertexes[i].values[4];
            args.freq := shader.DeformVertexes[i].values[5];
            nv := numOfVerts;
            for j := 0 to numOfVerts-1 do begin
              v := Data[j].normal;
              p := Data[j].Position; // buffer it
              off := (p.x+p.y+p.z) / waveSize;
              args.phase := startoff + off;
              inc := Eval(args);
              Data[j].Position.x := p.x + inc*v.x;
              Data[j].Position.y := p.y + inc*v.y;
              Data[j].Position.z := p.z + inc*v.z;
            end;
          end;
        DEFORMVERTEXES_NORMAL :
          begin
            args.func := enWave(trunc(shader.DeformVertexes[i].values[1]));
            args.base := shader.DeformVertexes[i].values[2];
            args.amp  := shader.DeformVertexes[i].values[3];
            args.freq := shader.DeformVertexes[i].values[4];
            nv := numOfVerts;
            for j := 0 to numOfVerts-1 do begin
              v := Data[j].normal;
              p := Data[j].Position; // buffer it
              off := (p.x+p.y+p.z) / (speed*waveSize);
              args.phase := startoff + off;
              inc := Eval(args);
              Data[j].Normal.x := p.x + inc*v.x;
              Data[j].Normal.y := p.y + inc*v.y;
              Data[j].Normal.z := p.z + inc*v.z;
            end;
          end;
        DEFORMVERTEXES_MOVE :
          begin
            args.func := enWave(trunc(shader.DeformVertexes[i].values[3]));
            args.base := shader.DeformVertexes[i].values[4];
            args.amp  := shader.DeformVertexes[i].values[5];
            args.phase := shader.DeformVertexes[i].values[6];
            args.freq := shader.DeformVertexes[i].values[7];
            for j := 0 to numOfVerts-1 do begin
              inc := Eval(args);
              Data[j].position.x := Data[j].position.x + inc*shader.DeformVertexes[i].values[0];
              Data[j].position.y := Data[j].position.y + inc*shader.DeformVertexes[i].values[1];
              Data[j].position.z := Data[j].position.z + inc*shader.DeformVertexes[i].values[2];
            end;
          end;
        DEFORMVERTEXES_AUTOSPRITE :
          begin
            (*
            glGetFloatv(GL_MODELVIEW_MATRIX, @cam_mat);
            right := Make3DVector_f(cam_mat[0], cam_mat[4], cam_mat[8]);
            up := Make3DVector_f(cam_mat[1], cam_mat[5], cam_mat[9]);
            right := vec_normalize(right);
            up := vec_normalize(up);
            middle := Make3DVector_f(0,0,0);
            // Compute middle of quad
            for j := 0 to numOfVerts-1 do begin
              middle.X := middle.X + Data[j].position.x;
              middle.Y := middle.Y + Data[j].position.y;
              middle.Z := middle.Z + Data[j].position.z;
            end;
            middle := vec_scale(middle, 1/numOfVerts);
            //glColor3ub(255, 0, 0);
            //glPointSize(2);
            // middle point
            //glBegin(GL_POINTS);
            //  glVertex3f(middle.X, middle.y, middle.Z);
            //glEnd;
            //glPointSize(4);
            //glColor3ub(0, 0, 255);
            //glBegin(GL_POINTS);
            // world points
            for j := 0 to numOfVerts-1 do begin
              p.x := middle.X + (Data[j].Position.X - middle.x);
              p.y := middle.y + (Data[j].Position.y - middle.y);
              p.z := middle.z + (Data[j].Position.z - middle.z);
           //   glVertex3f(p.X, p.y, p.Z);
            end;
            //glEnd;
            //glPointSize(8);
            //glColor3ub(255, 255, 0);
            //glBegin(GL_POINTS);
            // billboard points
            for j := 0 to numOfVerts-1 do begin
              dx := (Data[j].Position.X - middle.x);
              dy := (Data[j].Position.y - middle.y);
              dz := (Data[j].Position.z - middle.z);
{              if (dx < 0) or (dz < 0) then
                if (dx < 0) and (dz < 0) then
                  dx := -Sqrt(dx*dx + dz*dz)
                else
                  dx :=  Sqrt(dx*dx + dz*dz)
              else
                dx := Sqrt(dx*dx + dz*dz);}
              p := vec_add(vec_scale(up, dy), vec_scale(right, dx));
              p := vec_add(middle, p);
            //  glVertex3f(p.X, p.y, p.Z);
              Data[j].Position := p;
            end;
            //glEnd;
            //glColor3ub(255, 255, 255);
            *)
          end;
        DEFORMVERTEXES_BULGE :
          begin
            args.func := WF_SIN;
            args.base := 0;
            args.amp  := shader.DeformVertexes[i].values[1];
            args.freq := 1/(speed*shader.DeformVertexes[i].values[0]);
            nv := numOfVerts;
            for j := 0 to numOfVerts-1 do begin
              st := Data[j].TextureCoord;
              v := Data[j].normal;
              p := Data[j].Position; // buffer it
              args.phase := st.X*speed*shader.DeformVertexes[i].values[2];
              inc := Eval(args);
              Data[j].Position.x := p.x + inc*v.x;
              Data[j].Position.y := p.y + inc*v.y;
              Data[j].Position.z := p.z + inc*v.z;
            end;
          end;
      end;
    end;
  end;
end;

procedure TShaderManager.SetVertexColors_Shader(shaderId, stageId, numOfVerts : integer; var Data : array of TBSPVertex);
var shader : TShader;
    i, j : integer;
    c : integer;
begin
  shader := Items[shaderId];
  for j := 0 to numOfVerts-1 do begin
    colorBuffer[j, 0] := Data[j].Color[0]; // red
    colorBuffer[j, 1] := Data[j].Color[1]; // green
    colorBuffer[j, 2] := Data[j].Color[2]; // blue
    colorBuffer[j, 3] := Data[j].Color[3]; // gamma
  end;

  if shader.numOfLayers > 0 then begin
    if shader.Layers[stageId].numOfrgbGen > 0 then begin
      case shader.Layers[stageId].rgbGen[0].rgbGType of
        rgbGen_IdentityLightning :
          begin
            for j := 0 to numOfVerts-1 do begin
              colorBuffer[j, 0] := 127; // red
              colorBuffer[j, 1] := 127; // green
              colorBuffer[j, 2] := 127; // blue
              colorBuffer[j, 3] := 255; // gamma
            end;
          end;
        rgbGen_Identity :
          begin
            for j := 0 to numOfVerts-1 do begin
              colorBuffer[j, 0] := 255; // red
              colorBuffer[j, 1] := 255; // green
              colorBuffer[j, 2] := 255; // blue
              colorBuffer[j, 3] := 255; // gamma
            end;
          end;
        rgbGen_Wave :
          begin
            c := Round(255*Eval(shader.Layers[stageId].rgbGen[0].Wave));
            if c < 0 then c := 0;
            if c > 255 then c := 255;
            for j := 0 to numOfVerts-1 do begin
              colorBuffer[j, 0] := c;   // red
              colorBuffer[j, 1] := c;   // green
              colorBuffer[j, 2] := c;   // blue
              colorBuffer[j, 3] := 255; // gamma
            end;
          end;
       else begin
        for j := 0 to numOfVerts-1 do begin
          colorBuffer[j, 0] := Data[j].Color[0]; // red
          colorBuffer[j, 1] := Data[j].Color[1]; // green
          colorBuffer[j, 2] := Data[j].Color[2]; // blue
          colorBuffer[j, 3] := Data[j].Color[3]; // gamma
        end;
       end;
      end;
    end
    else begin
      // no rgbgen use vertex colors
      for j := 0 to numOfVerts-1 do begin
        colorBuffer[j, 0] := Data[j].Color[0]; // red
        colorBuffer[j, 1] := Data[j].Color[1]; // green
        colorBuffer[j, 2] := Data[j].Color[2]; // blue
        colorBuffer[j, 3] := Data[j].Color[3]; // gamma
      end;
    end;
  end;

  if shader.NonOpaque then
  for j := 0 to numOfVerts-1 do begin
  If colorBuffer[j, 0] > 100 then
    colorBuffer[j, 0] := 100; // red
  If colorBuffer[j, 1] > 100 then
    colorBuffer[j, 1] := 100; // green
  If colorBuffer[j, 2] > 100 then
    colorBuffer[j, 2] := 100; // blue
  If colorBuffer[j, 3] > 100 then
    colorBuffer[j, 3] := 100; // gamma
  end;

end;

procedure TShaderManager.SetTextureCoordinates_Shader(Const shaderId, stageId, numOfVerts : integer; var Data : array of TBSPVertex);
var i, j : integer;
    shader : TShader;
    dir : TVector3f;
    // tcmod
    angulo, cost, sint, factor : single;
    v, scroll : TVector2f;
    v2 : TVector3f;
    now : single;
    frame : integer;
begin
  shader := Items[shaderId];
(*  for j := 0 to numOfVerts-1 do
    texCoordBuffer[j] := Data[j].TextureCoord;
*)
  if shader.numOfLayers > 0 then begin
    //Animation
    if shader.Layers[StageID].animFreq <> 0 then begin
      frame := Trunc(gTimer.TotalTime *shader.Layers[StageID].animFreq) mod shader.Layers[StageID].numOfTextures;
      if frame >= shader.Layers[StageID].numOfTextures then
        frame := 0;
      Shader.Layers[StageID].curTexture := frame;
    end
    else
      Shader.Layers[StageID].curTexture := 0;
    // tcGen  ******************************************************************
    case shader.Layers[stageId].tcGen of
      TCGEN_LIGHTMAP :
        begin
          for j := 0 to numOfVerts-1 do
            Data[j].TextureCoord := Data[j].LightmapCoord;
        end;
      TCGEN_ENVIRONMENT :
        begin
          // wrong calc !!!!
          for j := 0 to numOfVerts-1 do begin
            dir := SubtractVector(Data[j].Position, {$IfDef HELIOS_APP}Camera.GetCurrentPosition{$Else}EnvironmentPos{$EndIf});//vec_sub(Data[j].Position, Camera.GetCurrentPosition);
            //dir := Normalize(dir);//vec_normalize(dir);
            Normalize(dir);
{            dir.x := dir.x + Data[j].Normal.X;
            dir.y := dir.y + Data[j].Normal.Y;
            texCoordBuffer[j].X := dir.x;
            texCoordBuffer[j].Y := dir.y;}
            dir.x := dir.x + Data[j].Normal.X;
            dir.z := dir.z + Data[j].Normal.z;
            Data[j].TextureCoord.X := dir.x;
            Data[j].TextureCoord.y := dir.z;

          end;

        end;
      else
        {for j := 0 to numOfVerts-1 do
          Data[j].TextureCoord := Data[j].TextureCoord};
    end;

    // tcMod *******************************************************************
    if shader.Layers[stageId].numOftcMods > 0 then begin
      for i := 0 to numOfVerts-1 do begin
        for j := 0 to shader.layers[stageId].numOftcMods-1 do begin
          case shader.Layers[stageId].tcMods[j].modType of
            TCMOD_ROTATE :
              begin
                angulo := shader.Layers[stageId].tcMods[j].Values[0]*gTimer.TotalTime/M_180_PI;
                cost := cos(angulo);
                sint := sin(angulo);
                v := Data[i].TextureCoord;
                v.x := v.x - 0.5;
                v.y := v.y - 0.5;
                Data[i].TextureCoord.x := v.x*cost - v.y*sint + 0.5;
                Data[i].TextureCoord.y := v.x*sint + v.y*cost + 0.5;
              end;
            TCMOD_SCALE :
              begin
                Data[i].TextureCoord.x := (Data[i].TextureCoord.x-0.5) * shader.Layers[stageId].tcMods[j].Values[0] + 0.5;
                Data[i].TextureCoord.y := (Data[i].TextureCoord.y-0.5) * shader.Layers[stageId].tcMods[j].Values[1] + 0.5;
              end;
            TCMOD_SCROLL :
              begin
                scroll.x := shader.Layers[stageId].tcMods[j].Values[0]*gTimer.TotalTime;
                scroll.y := shader.Layers[stageId].tcMods[j].Values[1]*gTimer.TotalTime;
                Data[i].TextureCoord.x := Data[i].TextureCoord.x + scroll.x;
                Data[i].TextureCoord.y := Data[i].TextureCoord.y - scroll.y;
              end;
            TCMOD_STRETCH :
              begin
                factor := (1/Eval(shader.Layers[StageID].tcMods[j].Wave));
                Data[i].TextureCoord.x := (Data[i].TextureCoord.x - 0.5) * factor + 0.5;
                Data[i].TextureCoord.y := (Data[i].TextureCoord.y - 0.5) * factor + 0.5;
              end;
            TCMOD_TRANSFORM :
              begin
                v := Data[i].TextureCoord;
                Data[i].TextureCoord.x := v.x*shader.Layers[stageId].tcMods[j].Values[0] + v.y*shader.Layers[stageId].tcMods[j].Values[2] + shader.Layers[stageId].tcMods[j].Values[4];
                Data[i].TextureCoord.y := v.x*shader.Layers[stageId].tcMods[j].Values[1] + v.y*shader.Layers[stageId].tcMods[j].Values[3] + shader.Layers[stageId].tcMods[j].Values[5];
              end;
            TCMOD_TURB :
              begin
//                v2 := Data[i].Position;
                v2.x := Data[i].TextureCoord.x;
                v2.y := Data[i].TextureCoord.y;
                v2.z := Data[i].Position.z;
                Data[i].TextureCoord.x := Data[i].TextureCoord.x + Sin((v2.x+v2.z)*0.1*shader.Layers[stageId].tcMods[j].Values[3] + gTimer.TotalTime + shader.Layers[stageId].tcMods[j].Values[2])*shader.Layers[stageId].tcMods[j].Values[1]*0.5;
                Data[i].TextureCoord.y := Data[i].TextureCoord.y + Sin((v2.y     )*0.1*shader.Layers[stageId].tcMods[j].Values[3] + gTimer.TotalTime + shader.Layers[stageId].tcMods[j].Values[2])*shader.Layers[stageId].tcMods[j].Values[1]*0.5;

//                v := texCoordBuffer[i];
//                Data[i].TextureCoord.x := Data[i].TextureCoord.x + Sin((v.x     )*0.1*shader.Layers[stageId].tcMods[j].Values[3] + gTimer.TotalTime + shader.Layers[stageId].tcMods[j].Values[2])*shader.Layers[stageId].tcMods[j].Values[1]*0.5;
//                Data[i].TextureCoord.y := Data[i].TextureCoord.y + Sin((v.y     )*0.1*shader.Layers[stageId].tcMods[j].Values[3] + gTimer.TotalTime + shader.Layers[stageId].tcMods[j].Values[2])*shader.Layers[stageId].tcMods[j].Values[1]*0.5;
              end;
          end;
        end;
      end;
    end;
  end;
end;

procedure TShaderManager.SetState(ShaderID, StageID : integer);
var shader :TShader;
begin
try             
  shader := Items[ShaderId];
  // Depth func
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(Items[ShaderID].Layers[StageID].DepthFunc);

  // Alpha
  if Items[ShaderID].Layers[StageID].doAlphaTest = True then begin
    glEnable(GL_ALPHA_TEST);
    glAlphaFunc(Items[ShaderID].Layers[StageID].AlphaFunc, Items[ShaderID].Layers[StageID].AlphaRef);
  end
  else
    glDisable(GL_ALPHA_TEST);

  if (Items[shaderId].Cull = GL_NONE) then
  glDisable(GL_CULL_FACE)
  else
  glEnable(GL_CULL_FACE);

  // Z-Buffer - needs sorted shaders
  if (Items[shaderId].Cull = GL_NONE) or (Items[shaderId].TotallyTrans) or (Items[ShaderID].Layers[StageID].doAlphaTest) then begin
    if Items[ShaderID].Layers[StageID].DepthWrite then
      glDepthMask(true)
    else
      glDepthMask(false);
  end
  else
    glDepthMask(true);

  // Blending
  if (Items[ShaderID].Layers[StageID].doBlending) then begin
    glEnable(GL_BLEND);
    glBlendFunc(Items[shaderId].Layers[stageId].SrcBlend, Items[shaderId].Layers[stageId].DestBlend);
//    If (Items[shaderId].Layers[stageId].SrcBlend = GL_DST_COLOR) and (Items[shaderId].Layers[stageId].SrcBlend = {GL_ONE_MINUS_DST_ALPHA}GL_ZERO) then
//    glDisable(GL_BLEND);
  end
  else
    glDisable(GL_BLEND);

  if Shader.Layers[StageID].TexClamp then begin
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  end
  else begin
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  end;

  if Shader.polygonOffset then
  glPolygonOffset(0.00001,100)
  else
  glPolygonOffset(0.1,100);

{  if Shader.polygonOffset then
  Cons.AddIfDifferent(Shader.Name);}

{  if Shader.Layers[StageID].tcGen = tcGen_Environment then begin
    glTexGenf(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glTexGenf(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
  end
  else begin
    glTexGenf(GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
    glTexGenf(GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
  end;
}
  except
//    ShowMessage('setStage');
  end;
end;

Procedure TShaderManager.DrawFaces(var Vertexs : array of TBSPVertex; Const CenterOfModel, Position : TVector3f; Vertex_No : Integer; ShaderId : Integer; UseNormals : Boolean; sTan,tTan : Array of TVector3f);
var
shader : TShader;
i,texId,v : Integer;
Light : TVector3f;
begin
If (ShaderId < 0) or (ShaderID >= Count) then exit;

Shader := Items[ShaderID];

Light := SubtractVector(LightPos,AddVector(CenterOfModel, Position));

// culling
 // if shader.Cull = GL_NONE then
 //   glDisable(GL_CULL_FACE);
 { else begin
    glEnable(GL_CULL_FACE);
    glCullFace(shader.cull);
  end;  }

//  for i := 0 to Vertex_No-1 do
//   inputBuffer[i] := Vertexs[i];

  SetVertexCoords_Shader(shaderId, Vertex_No,Vertexs);

glColor3f(1,1,1);

If (not UseBumpMaps) and (Shader.numOfLayers < 1) then
begin
inc(Shader.numOfLayers);
SetLength(Shader.Layers,Shader.numOfLayers);
Shader.Layers[0].TextureId[0] := Shader.DiffuseMapID;
Shader.Layers[0].numOfTextures := 1;
end;

If Shader.BumpMapID > 0 then
If UseBumpMaps then
begin

glBindProgramARB(GL_VERTEX_PROGRAM_ARB, BumpVP);
glEnable(GL_VERTEX_PROGRAM_ARB);
glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, -Light.X,-Light.Z,Light.Y,0);
glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 1, 0,0,0,0);

glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, BumpFP);
glEnable(GL_FRAGMENT_PROGRAM_ARB);
glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 0, 1/50,1/50,1/50,1/50); // invRadius
glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 1, 0,0,0,0);    // ambient

// Bind base map to texture unit 0
glActiveTexture(GL_TEXTURE0);
glEnable(GL_TEXTURE_2D);
glBindTexture(GL_TEXTURE_2D, Shader.DiffuseMapID);

// Bind normal map to texture unit 1
glActiveTexture(GL_TEXTURE1);
glBindTexture(GL_TEXTURE_2D, Shader.BumpMapID);
glEnable(GL_TEXTURE_2D);
// Bind specular map to texture unit 2
glActiveTexture(GL_TEXTURE2);
glBindTexture(GL_TEXTURE_2D, Shader.SpecularMapID);
glEnable(GL_TEXTURE_2D);

// Set vertex array
glVertexPointer(3, GL_FLOAT, SizeOf(TBSPVertex), @Vertexs[0].Position);
glEnableClientState(GL_VERTEX_ARRAY);
// Set normal array
glNormalPointer(GL_FLOAT, SizeOf(TBSPVertex), @Vertexs[0].Normal);
glEnableClientState(GL_NORMAL_ARRAY);
// Set texture coords array
glTexCoordPointer(2, GL_FLOAT, sizeof(TBSPVertex), @Vertexs[0].TextureCoord);
glEnableClientState(GL_TEXTURE_COORD_ARRAY);
// Send sTangent on tex coord 1
glClientActiveTexture(GL_TEXTURE1);
glTexCoordPointer(3, GL_FLOAT, sizeof(TVector3f), @sTan[0]);
glEnableClientState(GL_TEXTURE_COORD_ARRAY);
glClientActiveTexture(GL_TEXTURE0);
// Send tTangent on tex coord 2
glClientActiveTexture(GL_TEXTURE2);
glTexCoordPointer(3, GL_FLOAT, sizeof(TVector3f), @tTan[0]);
glEnableClientState(GL_TEXTURE_COORD_ARRAY);
glClientActiveTexture(GL_TEXTURE0);
// Send Norma on tex coord 3
glClientActiveTexture(GL_TEXTURE3);
glTexCoordPointer(3, GL_FLOAT, SizeOf(TBSPVertex), @Vertexs[0].Normal);
glEnableClientState(GL_TEXTURE_COORD_ARRAY);
glClientActiveTexture(GL_TEXTURE0);
// Render
glDrawArrays(GL_TRIANGLES,0,Vertex_No);

// Disable textures
glActiveTexture(GL_TEXTURE1);
glDisable(GL_TEXTURE_2D);
glActiveTexture(GL_TEXTURE2);
glDisable(GL_TEXTURE_2D);
glActiveTexture(GL_TEXTURE0);
//glDisable(GL_TEXTURE_2D);

glDisable(GL_VERTEX_PROGRAM_ARB);
glDisable(GL_FRAGMENT_PROGRAM_ARB);

exit;
end
else
begin

// Bind base map to texture unit 0
glActiveTexture(GL_TEXTURE0);
glEnable(GL_TEXTURE_2D);
glBindTexture(GL_TEXTURE_2D, Shader.DiffuseMapID);

// Set vertex array
glVertexPointer(3, GL_FLOAT, SizeOf(TBSPVertex), @Vertexs[0].Position);
glEnableClientState(GL_VERTEX_ARRAY);
// Set normal array
glNormalPointer(GL_FLOAT, SizeOf(TBSPVertex), @Vertexs[0].Normal);
glEnableClientState(GL_NORMAL_ARRAY);
// Set texture coords array
glTexCoordPointer(2, GL_FLOAT, sizeof(TBSPVertex), @Vertexs[0].TextureCoord);
glEnableClientState(GL_TEXTURE_COORD_ARRAY);

// Render
glDrawArrays(GL_TRIANGLES,0,Vertex_No);

exit;
end;


//  glVertexPointer(3, GL_FLOAT, sizeof(TBSPVertex), @inputBuffer[0].Position);

  if shader.numOfLayers > 0 then begin
    for i := 0 to shader.numOfLayers-1 do begin
    If not shader.Layers[i].Draw then continue; // Shader error!
      if (Items[shaderId].Layers[i].numOfrgbGen > 0) then begin
        SetVertexColors_Shader(shaderid,i,Vertex_no,Vertexs);
     //   SetVertexColors(shaderId, i, face.numMeshVerts);
     //   glColorPointer(4, GL_UNSIGNED_BYTE, 0, @colorBuffer[0]);
    //    glEnableClientState(GL_COLOR_ARRAY);
    glColor4f(colorBuffer[0][0],colorBuffer[0][1],colorBuffer[0][2],colorBuffer[0][3]);
      end
      else
        glColor3f(1,1,1);

      SetTextureCoordinates_Shader(shaderid,i,Vertex_no,Vertexs);

        texId := shader.Layers[i].TextureId[shader.Layers[i].curTexture];

      glEnable(GL_TEXTURE_2D);

      glBindTexture(GL_TEXTURE_2D, texId);
      texId := LastTexID;

      SetState(shaderId, i );
      //glDisable(GL_CULL_FACE);
      glBegin(GL_TRIANGLES);
      For v := 0 to Vertex_No-1 do
      begin
      If UseNormals then
      glNormal3f(Vertexs[v].Normal.X,Vertexs[v].Normal.Y,Vertexs[v].Normal.Z);
      glTexCoord2f(Vertexs[v].TextureCoord.X,Vertexs[v].TextureCoord.Y);
      glVertex3f(Vertexs[v].Position.X,Vertexs[v].Position.Y,Vertexs[v].Position.Z);
      end;
      glend;

    end;
  end;

  SetDefaults;
end;

function TShaderManager.GetTextureNameInShaderFormat(Texture : String) : string;
var
Tex : string;
begin

Tex := Texture;

if Length(ExtractFileExt(Tex)) > 0 then
Tex := copy(Tex,1,length(Tex)-Length(ExtractFileExt(Tex)));

Tex := StringReplace(Tex,'/','\',[rfReplaceAll, rfIgnoreCase]);

Result := Tex;

end;

Procedure TShaderManager.AddToShaderManager(Texture : string);
begin

AddRequiredShader(GetTextureNameInShaderFormat(Texture));

end;

Function TShaderManager.GetShaderId(Filename : String) : Integer;
begin
Result := -1;
Result := IndexOf(GetTextureNameInShaderFormat(Filename));
end;

Function GetD3BumpMapFromAddNormals(S : String) : String;
var
X,Y : Integer;
begin

If Pos('addnormals',Ansilowercase(S)) < 1 then
begin
Result := S;
exit;
end;

X := Pos('(',S);
Y := Pos(',',S);

Result := Copy(S,x+1,(y-1)-x);

end;

Procedure TShaderManager.ClearShaderErrors;
begin
ShaderErrors.ShaderError_No := 0;
SetLength(ShaderErrors.ShaderError,0);
end;

Procedure TShaderManager.AddShaderError(ShaderID : Cardinal; ShaderName, ErrorMessage : String; Layer : Cardinal);
var
SE : Integer;
begin
SE := ShaderErrors.ShaderError_No;
Inc(ShaderErrors.ShaderError_No);

SetLength(ShaderErrors.ShaderError,ShaderErrors.ShaderError_No);
ShaderErrors.ShaderError[SE].ShaderID := ShaderID;
ShaderErrors.ShaderError[SE].ShaderName := ShaderName;
ShaderErrors.ShaderError[SE].ErrorMessage := ErrorMessage;
ShaderErrors.ShaderError[SE].Layer := Layer;
end;


end.
