unit Helios_Sound;

{$I Global_Conditionals.INC}

interface

uses fmod, fmodtypes, fmoderrors, SysUtils, Math3d{, q3types};

type
  // sound lib entry
  TSample = record
    Name : string;
    SamplePtr : PFSoundSample;
    Loop : boolean;
    Use3D : boolean;
    //Position : TVector3f;
    //Volume : integer;
    //UsePosition : boolean;
  end;
  // emitters can share the same samples
  TEmitter = record
    SampleId : integer;
    SamplePtr : PFSoundSample;
    Position : TVector3f;
    UsePosition :boolean;
    //Loop : boolean;
  end;


Const
  MaxChannels = 300;//128; // = Max. emitters
  MaxSamples = 300;//256;  // = Max. entries in sound lib
  // 8 untits = 1 foot = 12 inches = 30,48cm
  // 8 units * 3.2 = 25.6 units = 1 m
  FullScreenRoqMode = FSOUND_STEREO or FSOUND_16BITS or FSOUND_2D;

type
  TSoundEngine = class(TObject)
  private
    FSoundLib : array[0..MaxSamples-1] of TSample;
    FnumOfSamples : integer;
    FnumOfEmitters : integer;
    FSampleRate : integer;
    FEnabled : boolean;
    FCameraChannel : integer;
    Procedure SetFMOD;
  protected
  public
    FEmitters : array[0..MaxChannels-1] of TEmitter;
    constructor Create;
    destructor Destroy; override;
    function NewSample(Mode : Cardinal) : integer;
    // returns index of FSoundLib
    function LoadSample(SampleID : integer; name : string; Loop, Use3D : boolean) : integer;
    // returns index of FEmitters ( = channel)
    function NewEmitter(EmitterId,SampleId : integer; Position : TVector3f; UsePosition, IsCamera : boolean) : integer;
    //procedure Play(Channel : integer; Pause : boolean);
    //procedure Update(Channel : integer; Position : TVector3f);
    procedure PlaySound(Channel : integer);
    procedure StopSound(Channel : integer);
    procedure Update(PlayerPosition : TVector3f);
    Procedure ClearSound;
    function CPU_Usage : single;
    Procedure Reload;
    Procedure StopALLSound;
    property Enabled : boolean read FEnabled;
    property Count : integer read FnumOfSamples;
  end;

  Function FX_Player_Sound_Add(FileName : String) : integer;
  Procedure FX_Player_Sound_Play(Snd : integer);
  Procedure SetMusic(Filename : string; Loop : Boolean);
  Procedure SetFXSound(Snd : Integer);
  Procedure StopFXSound;
  Function FXSoundPlaying : Boolean;
  Function LoadFXSample(Filename : String) : Integer;

implementation

uses {$IfDEF HELIOS_APP}Helios_Global,Helios_Filesystem,Helios_RTCW_SoundScriptSystem, Cam{$Else}Helios_Basic{$EndIf};

Procedure TSoundEngine.ClearSound;
begin
FnumOfSamples := 0;
  FnumOfEmitters := 0;
  FCameraChannel := -1;
end;

Procedure TSoundEngine.SetFMOD;
var caps: Cardinal;
    i : integer;
    enm: TFSoundOutputTypes;
begin
  FEnabled := false;
  FSampleRate := 44100;
  FnumOfSamples := 0;
  FnumOfEmitters := 0;
  FCameraChannel := -1;
  if FMOD_VERSION <= FSOUND_GetVersion then begin
    FSOUND_SetOutput(FSOUND_OUTPUT_DSOUND);

    enm := FSOUND_GetOutput();
    {$IfDef HELIOS_CONSOLE}
    if enm <> FSOUND_OUTPUT_DSOUND then begin
      Cons.AddMsg('Err: DirectSound Output was not successfully set!');
      case enm of
        FSOUND_OUTPUT_WINMM:    Cons.AddMsg('  Sound Output is now: Windows Multimedia Waveout');
        FSOUND_OUTPUT_DSOUND:   Cons.AddMsg('  Sound Output is now: Direct Sound - Dont believe what you read.');
        FSOUND_OUTPUT_A3D:      Cons.AddMsg('  Sound Output is now: A3D');
        FSOUND_OUTPUT_NOSOUND:  Cons.AddMsg('  Sound Output is now: No Sound - Thats usefull...');
      end;
    end;
    {$EndIf}

    {$IfDef HELIOS_CONSOLE}
    Cons.AddMsg('');
    Cons.AddMsg('--- Sound ---');
    {$EndIf}
    {Cons.AddMsg('--- DRIVER LIST ---');
    for i := 0 to FSOUND_GetNumDrivers() - 1 do begin
      Cons.AddMsg(inttostr(i + 1) + ' - ' + FSOUND_GetDriverName(i)); // print driver names
      FSOUND_GetDriverCaps(i, caps);
      if caps and FSOUND_CAPS_HARDWARE <> 0 then
        Cons.AddMsg('  - Driver supports hardware 3D sound!');
      if caps and FSOUND_CAPS_EAX2 <> 0 then
        Cons.AddMsg('  - Driver supports EAX 2 reverb!');
      if caps and FSOUND_CAPS_EAX3 <> 0 then
        Cons.AddMsg('  - Driver supports EAX 3 reverb!');
      if Caps = 0 then
        Cons.AddMsg('  - This sound card cannot support hardware 3D. Software mode will be used.');
    end;           }
    {$IfDef HELIOS_CONSOLE}Cons.AddMsg('Setting Driver to '+ FSound_GetDriverName(0));{$EndIf}
    FSOUND_SetDriver(0);
     FSOUND_SetMinHardwareChannels(MaxChannels);
    if FSOUND_Init(FSampleRate, MaxChannels, 0) then begin
      FEnabled := true;
      FSOUND_SetVolume(FSOUND_ALL, 0);
    end;
    {$IfDef HELIOS_CONSOLE}
    case (FSOUND_GetMixer()) of
      FSOUND_MIXER_BLENDMODE:     Cons.AddMsg('Mixer Mode: FSOUND_MIXER_BLENDMODE');
      FSOUND_MIXER_MMXP5:         Cons.AddMsg('Mixer Mode: FSOUND_MIXER_MMXP5');
      FSOUND_MIXER_MMXP6:         Cons.AddMsg('Mixer Mode: FSOUND_MIXER_MMXP6');
      FSOUND_MIXER_QUALITY_FPU:   Cons.AddMsg('Mixer Mode: FSOUND_MIXER_QUALITY_FPU');
      FSOUND_MIXER_QUALITY_MMXP5: Cons.AddMsg('Mixer Mode: FSOUND_MIXER_QUALITY_MMXP5');
      FSOUND_MIXER_QUALITY_MMXP6: Cons.AddMsg('Mixer Mode: FSOUND_MIXER_QUALITY_MMXP6');
    end;
    {$EndIf}
  end;
end;

constructor TSoundEngine.Create;
begin
  inherited Create;
  SetFMOD;
end;

destructor TSoundEngine.Destroy;
begin
  FSOUND_Close();


  inherited Destroy;
end;

Procedure TSoundEngine.Reload;
begin
FSOUND_Close();
SetFMOD;
end;

Procedure TSoundEngine.StopALLSound;
var
x : integer;
begin

if FnumOfEmitters > 0 then
for x := 0 to FnumOfEmitters-1 do
FSOUND_StopSound(x);
end;

//function TSoundEngine.LoadSample(name : string; Position : TVector3f; UsePosition, Loop : boolean) : integer;
function TSoundEngine.LoadSample(SampleID : integer; name : string; Loop, Use3D : boolean) : integer;
var sptr : PFSoundSample;
    i, Mode : integer;
begin
  if not FEnabled then begin
    result := -1;
    exit;
  end;

  name := LowerCase(name);
  name := StringReplace(name, '/', '\', [rfReplaceAll]); // remove linux style path delimiters

(*  if FnumOfSamples > 0 then begin
    for i := 0 to FNumOfSamples-1 do begin
      if (FSoundLib[i].Name = name) and (FSoundLib[i].Loop = Loop) and (FSoundLib[i].Use3D = Use3D)then begin
        result := i; // sample already loaded
        exit;
      end;
    end;
  end;
*)
  Mode := 0;
  if Loop then
    Mode := Mode + FSOUND_LOOP_NORMAL
  else
    Mode := Mode + FSOUND_LOOP_OFF;
  if Use3D then
    Mode := Mode + FSOUND_HW3D
  else
    Mode := Mode + FSOUND_2D;

  sptr := FSOUND_Sample_Load(FSOUND_FREE, PChar(name), Mode ,0,0);

  If (SampleID > -1) and (SampleID < FnumOfSamples) then
  begin

  if sptr <> nil then begin
    FSoundLib[SampleID].Name := name;
    FSoundLib[SampleID].SamplePtr := sptr;
    FSoundLib[SampleID].Loop := Loop;
    FSoundLib[SampleID].Use3D := Use3D;
    result := SampleID;
    if Use3D then
      FSOUND_Sample_SetMinMaxDistance(sptr, 1, 1);

  end
  else
    result := -1;

  end
  else
  If FnumOfSamples < MaxSamples then
  begin

  if sptr <> nil then begin
    FSoundLib[FnumOfSamples].Name := name;
    FSoundLib[FnumOfSamples].SamplePtr := sptr;
    FSoundLib[FnumOfSamples].Loop := Loop;
    FSoundLib[FnumOfSamples].Use3D := Use3D;
    result := FnumOfSamples;
    inc(FnumOfSamples);
    if Use3D then
      FSOUND_Sample_SetMinMaxDistance(sptr, 1, 1);
  end
  else
    result := -1;

  end;

  // this section is evaluated only if the sample isn't loaded
(*
  if Loop then
    Mode := FSOUND_LOOP_NORMAL
  else
    Mode := 0;
  if UsePosition then begin
    Mode := Mode+FSOUND_HW3D;
    //Mode := Mode + FSOUND_2D;
  end
  else
    Mode := Mode + FSOUND_2D;
//  if Loop then
//    Mode := Mode + FSOUND_LOOP_NORMAL;
  sptr := FSOUND_Sample_Load(FSOUND_FREE, PChar(name), Mode ,0,0);
  if UsePosition then
    FSOUND_Sample_SetMinMaxDistance(sptr, 1, 1);
  if sptr <> nil then begin
    position.Z := - Position.Z;
    FSampleList[FnumOfSamples].Name := name;
    FSampleList[FnumOfSamples].SamplePtr := sptr;
    FSampleList[FnumOfSamples].Position  := Position;
    FSampleList[FnumOfSamples].UsePosition  := UsePosition;
    //FSampleList[FnumOfSamples].Volume := Volume;

    result := FnumOfSamples;
    inc(FnumOfSamples);
  end
  else
    result := -1;
*)
end;

function TSoundEngine.NewSample(Mode : Cardinal)  : integer;
var sptr : PFSoundSample;
begin
  if not FEnabled then begin
    result := -1;
    exit;
  end;
                                          // FSOUND_STEREO | FSOUND_16BITS, 44100, 255, 128, 255);
                                          // Original: 22050
  sptr := FSOUND_Sample_Alloc(FnumOfSamples,65536,Mode,21000,255,128,255);

  If FnumOfSamples < MaxSamples then
  begin

  if sptr <> nil then begin
    FSoundLib[FnumOfSamples].Name := 'NewSample';
    FSoundLib[FnumOfSamples].SamplePtr := sptr;
    FSoundLib[FnumOfSamples].Loop := False;
    FSoundLib[FnumOfSamples].Use3D := False;
    result := FnumOfSamples;
    inc(FnumOfSamples);
  end
  else
    result := -1;

  end;

end;

function TSoundEngine.NewEmitter(EmitterId,SampleId : integer; Position : TVector3f; UsePosition, IsCamera : boolean) : integer;
begin
  if not FEnabled then exit;

  result := -1;

  If (EmitterId > -1) and (EmitterId < FnumOfEmitters) then
  if (SampleId >= 0) and (SampleId < FnumOfSamples) then
  begin
      FEmitters[EmitterId].SampleId := SampleId;
      FEmitters[EmitterId].SamplePtr := FSoundLib[SampleId].SamplePtr;
      FEmitters[EmitterId].Position := Position;
      FEmitters[EmitterId].UsePosition := UsePosition;
//      FSOUND_SetVolume(EmitterId, 0);
      if not IsCamera then
      FSOUND_PlaySoundEx(EmitterId, FEmitters[EmitterId].SamplePtr, nil, true);
      result := EmitterId;
      if IsCamera then
        FCameraChannel := EmitterId;
  exit;

  end;

  if (FnumOfEmitters < (MaxChannels-1)) then begin
    if (SampleId >= 0) and (SampleId < FnumOfSamples) then begin
      FEmitters[FnumOfEmitters].SampleId := SampleId;
      FEmitters[FnumOfEmitters].SamplePtr := FSoundLib[SampleId].SamplePtr;
      FEmitters[FnumOfEmitters].Position := Position;
      FEmitters[FnumOfEmitters].UsePosition := UsePosition;
//      FSOUND_SetVolume(FnumOfEmitters, 0);
      if not IsCamera then
      FSOUND_PlaySoundEx(FnumOfEmitters, FEmitters[FnumOfEmitters].SamplePtr, nil, true);
      result := FnumOfEmitters;
      if IsCamera then
        FCameraChannel := FnumOfEmitters;
      inc(FnumOfEmitters);
    end;
  end;

(*  if (Channel >= 0) and (Channel < FnumOfSamples) then begin
    FSOUND_PlaySoundEx(Channel, FSampleList[Channel].SamplePtr, nil, Pause);
    FSOUND_SetVolume(Channel, 0);
  end;
*)
end;

procedure TSoundEngine.PlaySound(Channel : integer);
begin
FSOUND_PlaySoundEx(Channel, FEmitters[Channel].SamplePtr, nil, false);
end;

procedure TSoundEngine.StopSound(Channel : integer);
begin
FSOUND_StopSound(Channel);
end;

(*
procedure TSoundEngine.Update(Channel : integer; Position : TVector3f);
var vel : TVector3f;
begin
  vel.x := 0;
  vel.y := 0;
  vel.z := 0;
  FSOUND_3D_SetAttributes(Channel, @Position, @vel);
  //FSOUND_3D_Listener_SetAttributes
  //FSOUND_3D_Update();
end;
*)

Function Max(V1,V2 : integer) : Integer;
begin

If V1 > V2 then
Result := V1
else
Result := V2;

end;

procedure TSoundEngine.Update(PlayerPosition : TVector3f);
var i : integer;
    vel, pos, point : TVector3f;
    distance : single;
    vol_factor : single;

begin
  if FEnabled and (FnumOfSamples > 0) then begin
    vol_factor := 255 / FX_VOL;
    vel.x := 0; vel.y := 0; vel.z := 0;
    pos := PlayerPosition;
    //@@pos.Z := -pos.z;
    //@@FSOUND_3D_Listener_SetAttributes(@pos, @vel, camera.eyedir.X, camera.eyedir.Y, -camera.eyedir.Z, 0, 1, 0);
    FSOUND_3D_Listener_SetAttributes(@pos, @vel, {$IfDEF HELIOS_APP}camera.eyedir.X, camera.eyedir.Y, camera.eyedir.Z{$Else}0,0,0{$EndIf}, 0, 1, 0);
    for i := 0 to FnumOfEmitters-1 do begin
      If Pause then
      begin
      FSOUND_SetPaused(i, True);
      Continue;
      end
      else
      FSOUND_SetPaused(i, false);

      // Distance -> volume
      if FEmitters[i].UsePosition then begin
        Point.x := FEmitters[i].Position.x  - PlayerPosition.x;
        Point.y := FEmitters[i].Position.y  - PlayerPosition.y;
        Point.z := FEmitters[i].Position.z  - PlayerPosition.z;
        distance := sqrt(Point.x*Point.x + Point.y*Point.y + Point.z*Point.z);
        if distance < 0 then
          distance := abs(distance);
        if distance <= 1024{512} then begin
          FSOUND_SetVolume(i, Max(Trunc((255 - Round(distance/4))/vol_factor),0));
          //FSOUND_3D_SetAttributes(i, @FSampleList[i].Position, @vel);
        end
        else
          FSOUND_SetVolume(i, 0);
      end
      else begin
        if (FCameraChannel = i) then begin
          FSOUND_SetVolume(i, FX_VOL);
          {$IfDEF HELIOS_APP}
          if Camera.PlaySound then begin
            if GroundType = 'NORMAL' then
              FSOUND_PlaySoundEx(FCameraChannel, FSoundLib[Camera.NormalSound+Camera.sndCnt].SamplePtr, nil, false)
            else if GroundType = 'METAL' then
              FSOUND_PlaySoundEx(FCameraChannel, FSoundLib[Camera.MetalSound+Camera.sndCnt].SamplePtr, nil, false)
            else if GroundType = 'WATER' then
              FSOUND_PlaySoundEx(FCameraChannel, FSoundLib[Camera.WaterSound+Camera.sndCnt].SamplePtr, nil, false);
(*            inc(Camera.sndCnt);
            if Camera.sndCnt > 3 then
              Camera.sndCnt := 0;*)
            Camera.sndCnt := Random(3);
            Camera.PlaySound := false;
          end;
          {$EndIf}
        end
        else
        begin

        If i <> MUSIC_Channel then // not music channel
        begin
        FSOUND_SetVolume(i, FX_VOL);
        Continue;
        end;

          FSOUND_SetVolume(i, MUSIC_VOL);
          if (FSOUND_GetCurrentPosition(i) = 0) and (MUSIC_Start = false) and (MUSIC_Loop = true) then
          begin
          MUSIC_Loop := false;
          //cons.AddMsg('sound should have changed to loop');
          FEmitters[i].SampleId := MUSIC_SND2;
          FEmitters[i].SamplePtr := FSoundLib[MUSIC_SND2].SamplePtr;
          FSOUND_PlaySoundEx(i, FEmitters[i].SamplePtr, nil, true);
          end;

          if (FSOUND_GetCurrentPosition(i) > 0) and (MUSIC_Start = true) then
          MUSIC_Start := false;

        end;             
      end;
     // FSOUND_Update(); //Doesn't seem to be needed, dramaticaly improves fps in most areas when removed.

    end;
  end;
end;

function TSoundEngine.CPU_Usage : single;
begin
  if FEnabled then
    result := FSOUND_GetCPUUsage
  else
    result := 0;
end;

Function FX_Player_Sound_Add(FileName : String) : integer;
var
FN : String;
begin
FN := DoesFileExist('\sound\player\'+{Player_Model+}'\' +Filename);

if FN = '' then
Exit;

inc(FX_Player_Sound_No);

SetLength(FX_Player_Sound,FX_Player_Sound_No);

FX_Player_Sound[FX_Player_Sound_No-1] := Sound.LoadSample(-1,FN, False, True);

Result := FX_Player_Sound_No-1; 
end;

Procedure FX_Player_Sound_Play(Snd : integer);
begin

if FX_Player_Sound_Channel = -1 then
begin
FX_Player_Sound_Channel := Sound.NewEmitter(-1,snd, {$IfDEF HELIOS_APP}Camera.GetCurrentPosition{$Else}SetVector(0,0,0){$EndIf}, True, false);
end
else
begin
Sound.FEmitters[FX_Player_Sound_Channel].Position := {$IfDEF HELIOS_APP}Camera.GetCurrentPosition{$Else}SetVector(0,0,0){$EndIf};
Sound.FEmitters[FX_Player_Sound_Channel].SampleId := FX_Player_Sound[Snd];
Sound.FEmitters[FX_Player_Sound_Channel].SamplePtr := Sound.FSoundLib[FX_Player_Sound[Snd]].SamplePtr;

FSOUND_PlaySoundEx(FX_Player_Sound_Channel, Sound.FEmitters[FX_Player_Sound_Channel].SamplePtr, nil, true);
end;
end;

Procedure SetMusic(Filename : string; Loop : Boolean);
var
FN : String;
begin

FN := DoesFileExist(Filename);

if FN = '' then
FN := DoesFileExist(Filename + '.wav');

if FN = '' then
FN := DoesFileExist(Filename + '.mp3');

if FN = '' then
exit;

if MUSIC_Channel <> -1 then
Sound.StopSound(MUSIC_Channel);

MUSIC_Loop := false;

MUSIC_SND1 := Sound.LoadSample(MUSIC_SND1,FN, Loop, false);
MUSIC_Channel := Sound.NewEmitter(MUSIC_Channel,MUSIC_SND1, SetVector(0,0,0), false, false);
end;

Function LoadFXSample(Filename : String) : Integer;
var
FN,FN2 : String;
begin
{$IfDEF HELIOS_APP}
FN2 := RTCWSoundSystem_GetSound(Filename);

FN := DoesFileExist(FN2);

if FN = '' then
FN := DoesFileExist(FN2 + '.wav');

if FN = '' then
FN := DoesFileExist(FN2 + '.mp3');

if FN = '' then
begin
{$IfDef HELIOS_CONSOLE}Cons.AddMsg('Couldn`t Find: ' + Filename);{$EndIf}
exit;
end;

FXS := Sound.LoadSample(FXS,FN, false, false);
//Cons.AddMsg('Loaded: ' + Filename + ' to: ' + inttostr(FXS));
{$EndIf}
end;

Procedure SetFXSound(Snd : Integer);
begin
If FXS = -1 then exit;
//Cons.AddMsg('Playing: ' + inttostr(FXS));
FX_Sound_Channel := Sound.NewEmitter(FX_Sound_Channel,FXS, SetVector(0,0,0), false, false);
end;

Procedure StopFXSound;
begin
if FX_Sound_Channel <> -1 then
Sound.StopSound(FX_Sound_Channel);
end;

Function FXSoundPlaying : Boolean;
begin
Result := False;

If FX_Sound_Channel > -1 then
If FSOUND_GetCurrentPosition(FX_Sound_Channel) > 0 then
Result := True;

end;

end.
