{
  RoQ Video File Decoding Routines

  Id Software's RoQ video file format decoder
  Dr. Tim Ferguson, 2001.
  Tim Ferguson: http://www.csse.monash.edu.au/~timf/

  Conversion by Osman Turan (C++ -> Delphi)
  Some codes rewritten

  This unit is a part of Quantum Engine or shortly QNE.
  Quantum Engine has being developed by Osman Turan
}

unit qne_RoQ;

interface

uses Classes, Math3d, SysUtils;

const
  ROQ_HEADER_ID1        = $1084;
  ROQ_HEADER_ID2        = $FFFFFFFF;
  ROQ_INFO              = $1001;
  ROQ_QUAD_CODEBOOK     = $1002;
  ROQ_QUAD_VQ           = $1011;
  ROQ_SOUND_MONO        = $1020;
  ROQ_SOUND_STEREO      = $1021;

  ROQ_CODE_MOT          = $00;
  ROQ_CODE_FCC          = $01;
  ROQ_CODE_SLD          = $02;
  ROQ_CODE_CCC          = $03;

  MAX_ROQ_CODEBOOK_SIZE = 256;
  ROQ_EMPTY_CODEWORD    = $FFFF0000;

type
  // codebook entry for 2x2 vector
  PRoQ_V2_Codebook = ^TRoQ_V2_Codebook;
  TRoQ_V2_Codebook = packed record
    y: array[0..3] of Byte;
    u, v: Byte;
  end;

  // codebook entry for 4x4 vector
  TRoQ_V4_Codebook = packed record
    v2_index: array[0..3] of Byte;
  end;

  TRoQChunkHeader = packed record
    id : Word;
    size : Cardinal;
    arg : Word;
  End;

  TRoQVideoChunk = packed record
    width, height : Word;
    Unknown1, Unknown2 : Word;
  End;

  PCinematic = ^TCinematic;
  TCinematic = packed record
    FrameBuffer: array[0..1] of PByteArray;
    Width, Height: Cardinal;
    RowStride: Cardinal;
    FPS: Word;
    CurrentFrame: Cardinal;
    StartTime: Double;

    fp: TStream;
    Chunk: TRoQChunkHeader;
    V2Codebook: array[0..MAX_ROQ_CODEBOOK_SIZE-1] of TRoQ_V2_Codebook;
    V4CodeBook: array[0..MAX_ROQ_CODEBOOK_SIZE-1] of TRoQ_V4_Codebook;
    SkipAudio: Boolean;

    AudioChannels: Word;
    AudioBuffer: PSmallIntArray;
    AudioSampleOffset: Cardinal;
    AudioSampleCount: Cardinal;

    // Stucuk Edit
    State : Cardinal; // 0 Stopped, 1 Playing, 2 Paused
    Texture : Cardinal; // OpenGL Texture We Will Render To
    KeepOpen : Boolean; // Defultly True, if False the roq will close at the end
    FmodEmitterID : Integer;
    RenderToScreen : Boolean; // Is it a "movie" if so we render it to screen in Helios_Roq_Engine
    NewSound : Boolean;
    Loop : Boolean;
    Filename : String;
    //
  end;

function RoQ_Open(const Filename: String;
                  const SkipAudio: Boolean;
                  const SampleCount: Cardinal): PCinematic;
procedure RoQ_Restart(cin: PCinematic);
procedure RoQ_Close(var cin: PCinematic);
function RoQ_ReadNextFrame(cin: PCinematic): Boolean;

implementation

uses qne_RoQTables;

//-------------------------------------------------------------------------- */
function GetGreenCoeff(const Value: Integer): Integer;
asm
        SAR     EAX, $10 // signed divide by 2^16
end;

procedure decode_block_4x4(dst0, dst1: PByteArray;
                           const y: PByteArray;
                           const u, v: Integer);
var r_coeff, g_coeff, b_coeff: Integer;
begin
  r_coeff := YUV_CR_R_TABLE[v];
  g_coeff := GetGreenCoeff(YUV_CB_G_TABLE[u] + YUV_CR_G_TABLE[v]);
  b_coeff := YUV_CB_B_TABLE[u];

  dst0[0] := YUV_CLIP_TABLE[y[0] + r_coeff];
  dst0[1] := YUV_CLIP_TABLE[y[0] + g_coeff];
  dst0[2] := YUV_CLIP_TABLE[y[0] + b_coeff];

  dst0[3] := YUV_CLIP_TABLE[y[1] + r_coeff];
  dst0[4] := YUV_CLIP_TABLE[y[1] + g_coeff];
  dst0[5] := YUV_CLIP_TABLE[y[1] + b_coeff];

  dst1[0] := YUV_CLIP_TABLE[y[2] + r_coeff];
  dst1[1] := YUV_CLIP_TABLE[y[2] + g_coeff];
  dst1[2] := YUV_CLIP_TABLE[y[2] + b_coeff];

  dst1[3] := YUV_CLIP_TABLE[y[3] + r_coeff];
  dst1[4] := YUV_CLIP_TABLE[y[3] + g_coeff];
  dst1[5] := YUV_CLIP_TABLE[y[3] + b_coeff];
end;

procedure decode_block_8x8(dst0, dst1: PByteArray;
                           const y: PByteArray;
                           const u, v: Integer);
var r_coeff, g_coeff, b_coeff: Integer;
begin
  r_coeff := YUV_CR_R_TABLE[v];
  g_coeff := GetGreenCoeff(YUV_CB_G_TABLE[u] + YUV_CR_G_TABLE[v]);
  b_coeff := YUV_CB_B_TABLE[u];

  dst0[ 0] := YUV_CLIP_TABLE[y[0] + r_coeff];
  dst0[ 1] := YUV_CLIP_TABLE[y[0] + g_coeff];
  dst0[ 2] := YUV_CLIP_TABLE[y[0] + b_coeff];

  dst0[ 3] := YUV_CLIP_TABLE[y[1] + r_coeff];
  dst0[ 4] := YUV_CLIP_TABLE[y[1] + g_coeff];
  dst0[ 5] := YUV_CLIP_TABLE[y[1] + b_coeff];

  dst0[ 6] := YUV_CLIP_TABLE[y[2] + r_coeff];
  dst0[ 7] := YUV_CLIP_TABLE[y[2] + g_coeff];
  dst0[ 8] := YUV_CLIP_TABLE[y[2] + b_coeff];

  dst0[ 9] := YUV_CLIP_TABLE[y[3] + r_coeff];
  dst0[10] := YUV_CLIP_TABLE[y[3] + g_coeff];
  dst0[11] := YUV_CLIP_TABLE[y[3] + b_coeff];

  PCardinal(@dst1[0])^ := PCardinal(@dst0[0])^;
  PCardinal(@dst1[4])^ := PCardinal(@dst0[4])^;
  PCardinal(@dst1[8])^ := PCardinal(@dst0[8])^;
end;

//-------------------------------------------------------------------------- */
procedure apply_vector_2x2(cin: PCinematic; x, y: Cardinal; cell : PRoQ_V2_Codebook);
var dst0, dst1: PByteArray;
begin
  dst0 := @cin.FrameBuffer[0, (y * cin.Width + x) * 3];
  dst1 := @dst0^[cin.RowStride];

  decode_block_4x4(dst0, dst1, @cell.y[0], cell.u, cell.v);
End;

//-------------------------------------------------------------------------- */
procedure apply_vector_4x4(cin: PCinematic; x, y: Cardinal; cell: PRoQ_V2_Codebook);
var
  dst0, dst1: PByteArray;
  p: array[0..3] of Byte;
begin
  p[0] := cell.y[0];
  p[1] := cell.y[0];
  p[2] := cell.y[1];
  p[3] := cell.y[1];

  dst0 := @cin.FrameBuffer[0, (y * cin.Width + x) * 3];
  dst1 := @dst0^[cin.RowStride];
  decode_block_8x8(dst0, dst1, @p[0], cell.u, cell.v);

  p[0] := cell.y[2];
  p[1] := cell.y[2];
  p[2] := cell.y[3];
  p[3] := cell.y[3];

  dst0 := @dst0^[cin.RowStride * 2];
  dst1 := @dst1^[cin.RowStride * 2];
  decode_block_8x8(dst0, dst1, @p[0], cell.u, cell.v);
End;

//-------------------------------------------------------------------------- */
procedure apply_motion_4x4(cin: PCinematic; x, y: Cardinal; mv: Byte; mean_x, mean_y: ShortInt);
Var
  mx, my: Cardinal;
  src, dst: PCardinalArray;
  i: Integer;
Begin
  mx := x + 8 - (mv shr 4) - mean_x;
  my := y + 8 - (mv and $0F) - mean_y;
  src := @cin.FrameBuffer[1, (my * cin.Width + mx) * 3];
  dst := @cin.FrameBuffer[0, (y * cin.Width + x) * 3];

  for i:=0 to 3 do
  begin
    dst^[0] := src^[0];
    dst^[1] := src^[1];
    dst^[2] := src^[2];

    src := Pointer(Cardinal(src) + cin.RowStride);
    dst := Pointer(Cardinal(dst) + cin.RowStride);
  end;
end;

//-------------------------------------------------------------------------- */
procedure apply_motion_8x8(cin: PCinematic; x, y: Cardinal; mv: Byte; mean_x, mean_y: ShortInt);
Var
  mx, my: Cardinal;
  src, dst: PCardinalArray;
  i: Integer;
begin
  mx := x + 8 - (mv shr 4) - mean_x;
  my := y + 8 - (mv and $0F) - mean_y;

  src := @cin.FrameBuffer[1, (my * cin.Width + mx) * 3];
  dst := @cin.FrameBuffer[0, (y * cin.Width + x) * 3];

  for i:=0 to 7 do
  begin
    dst^[0] := src^[0];
    dst^[1] := src^[1];
    dst^[2] := src^[2];
    dst^[3] := src^[3];
    dst^[4] := src^[4];
    dst^[5] := src^[5];

    src := Pointer(Cardinal(src) + cin.RowStride);
    dst := Pointer(Cardinal(dst) + cin.RowStride);
  end;
end;

//-------------------------------------------------------------------------- */
procedure RoQ_ReadChunk(cin: PCinematic);
begin
// just confidence clear all data
  FillChar(cin.Chunk, SizeOf(TRoQChunkHeader), 0);
  cin.fp.Read(cin.Chunk, SizeOf(TRoQChunkHeader));
end;

procedure RoQ_SkipBlock(cin: PCinematic; Size: Integer);
begin
  cin.fp.Seek(Size, soFromCurrent);
end;

procedure RoQ_SkipChunk(cin: PCinematic);
begin
  RoQ_SkipBlock(cin, cin.Chunk.size);
end;

function RoQ_Open(const Filename: String;
                  const SkipAudio: Boolean;
                  const SampleCount: Cardinal): PCinematic;
Var
  fp : TStream;
  cin: PCinematic;
  vidinfo: TRoQVideoChunk;
Begin
  Result := nil;

  fp := TFileStream.Create(Filename, fmOpenRead or fmShareDenyWrite);

  If fp = nil Then
    Exit;

  GetMem(cin, SizeOf(TCinematic));
  FillChar(cin^, SizeOf(TCinematic), 0);

  cin.fp := fp;

  RoQ_ReadChunk(cin);

  if not ((cin.Chunk.id = ROQ_HEADER_ID1) and
          (cin.Chunk.size = ROQ_HEADER_ID2)) then
    Exit;

  cin.FPS := cin.Chunk.arg;

// search video info chunk
  while cin.fp.Position < cin.fp.Size do
  begin
    RoQ_ReadChunk(cin);
    if cin.Chunk.id = RoQ_INFO then
    begin
      cin.fp.Read(vidinfo, SizeOf(TRoQVideoChunk));
      cin.Width := vidinfo.Width;
      cin.Height := vidinfo.Height;
      cin.RowStride := vidinfo.width * 3;
      Break;
    end
    else
      RoQ_SkipChunk(cin);
  end;

// we found it?
  if (cin.Width = 0) or (cin.Height = 0) then
    Exit;

  GetMem(cin.FrameBuffer[0], cin.Width * cin.Height * 3);
  GetMem(cin.FrameBuffer[1], cin.Width * cin.Height * 3);

  cin.SkipAudio := SkipAudio;
  if not cin.SkipAudio then
  begin
    cin.AudioSampleCount := SampleCount;
    GetMem(cin.AudioBuffer, cin.AudioSampleCount * SizeOf(SmallInt));
    // Fill with silence value (for 16-bit sound = 0);
    FillChar(cin.AudioBuffer^, cin.AudioSampleCount * SizeOf(SmallInt), 0);
  end;

  Cin.KeepOpen := True;
  Cin.Loop := False;
  Cin.FmodEmitterID := -1;
  RoQ_Restart(cin);

  Result := cin;
End;

procedure RoQ_Restart(cin: PCinematic);
begin
// restart stream, but ignore RoQ header
  cin.fp.Seek(SizeOf(TRoQChunkHeader), soFromBeginning);
  cin.CurrentFrame := 0;
  cin.State := 0;
  Cin.StartTime := 0;
end;

//-------------------------------------------------------------------------- */
procedure RoQ_Close(var cin: PCinematic);
begin
  if not Assigned(cin) then Exit;

{  cin.fp.Free;

  if Assigned(cin.FrameBuffer[0]) then
    FreeMem(cin.FrameBuffer[0]);

  if Assigned(cin.FrameBuffer[1]) then
    FreeMem(cin.FrameBuffer[1]);
                   }
  FreeMem(cin);
  cin := nil;
End;

procedure RoQ_ReadAudio(cin: PCinematic);
var
  snd_left, snd_right: SmallInt;
  buf : PByteArray;
  sample: Cardinal;
begin
cin.AudioSampleOffset := 0;

  if cin.SkipAudio then
  begin
    RoQ_SkipChunk(cin);
    Exit;
  end;

  case cin.Chunk.id of
    ROQ_SOUND_MONO:
      begin
        cin.NewSound := True;
        cin.AudioChannels := 1;
        snd_left := (cin.Chunk.arg);
        GetMem(buf, cin.Chunk.size);
        cin.fp.Read(buf[0], cin.Chunk.size);
        sample := 0;
        while sample < cin.Chunk.size do
        begin
          Inc(snd_left, ROQ_DPCM_TABLE[buf[sample]]);
          cin.AudioBuffer[cin.AudioSampleOffset] := snd_left;

          Inc(cin.AudioSampleOffset);
          if cin.AudioSampleOffset >= cin.AudioSampleCount then cin.AudioSampleOffset := 0;
          Inc(sample);

          // Stucuk
          // Make it like sterio for FMOD (so we don't need any convertions)
          cin.AudioBuffer[cin.AudioSampleOffset] := snd_left;

          Inc(cin.AudioSampleOffset);
          if cin.AudioSampleOffset >= cin.AudioSampleCount then cin.AudioSampleOffset := 0;
        end;

        FreeMem(buf);
      end;

    RoQ_SOUND_STEREO:
      begin
        cin.NewSound := True;
        cin.AudioChannels := 2;
        snd_left := (cin.Chunk.arg and $FF00);
        snd_right := ((cin.Chunk.arg and $FF) shl 8);
        GetMem(buf, cin.Chunk.size);
        cin.fp.Read(buf[0], cin.Chunk.size);

        sample := 0;
        while sample < cin.Chunk.size do
        begin
          Inc(snd_left, ROQ_DPCM_TABLE[buf[sample]]);

          cin.AudioBuffer[cin.AudioSampleOffset] := snd_left;

          Inc(cin.AudioSampleOffset);
          if cin.AudioSampleOffset >= cin.AudioSampleCount then cin.AudioSampleOffset := 0;

          Inc(sample);

          Inc(snd_right, ROQ_DPCM_TABLE[buf[sample]]);
          cin.AudioBuffer[cin.AudioSampleOffset] := snd_right;

          Inc(cin.AudioSampleOffset);
          if cin.AudioSampleOffset >= cin.AudioSampleCount then cin.AudioSampleOffset := 0;

      {    cin.AudioBuffer[cin.AudioSampleOffset] := snd_left;
          Inc(cin.AudioSampleOffset);
          if cin.AudioSampleOffset >= cin.AudioSampleCount then cin.AudioSampleOffset := 0;

          cin.AudioBuffer[cin.AudioSampleOffset] := snd_right;
          Inc(cin.AudioSampleOffset);
          if cin.AudioSampleOffset >= cin.AudioSampleCount then cin.AudioSampleOffset := 0;}

          Inc(sample);
        end;

        FreeMem(buf);
      end;
  else
    Exit;
  end;
End;

procedure RoQ_ReadVideo(cin: PCinematic);
var
  k, vqflg, vqflg_pos, vqid: Integer;
  bpos, xpos, ypos, x, y, xp, yp: Cardinal;
  tp, buf : PByteArray;
  mean_motion_y, mean_motion_x: Word;
  v4: TRoQ_V4_Codebook;
begin
  GetMem(buf, cin.Chunk.size);
  cin.fp.Read(buf^, cin.Chunk.size);

  bpos := 0;
  xpos := 0;
  ypos := 0;
  vqflg := 0;
  vqflg_pos := -1;

  mean_motion_y := ShortInt(cin.Chunk.arg and $FF);
  mean_motion_x := ShortInt((cin.Chunk.arg shr 8) and $FF);

  while (bpos < cin.Chunk.size) do
  begin
    yp := ypos;
    while (yp < ypos + 16) do
    begin
      xp := xpos;
      while (xp < xpos + 16) do
      begin
        if (vqflg_pos < 0) then
        begin
          vqflg := PWord(@buf[bpos])^;
          Inc(bpos, 2);
          vqflg_pos := 7;
        end;

        vqid := (vqflg shr (vqflg_pos * 2)) and $03;
        Dec(vqflg_pos);

        case vqid of
          ROQ_CODE_MOT : begin end;
          ROQ_CODE_FCC :
            begin
              apply_motion_8x8(cin, xp, yp, buf[bpos], mean_motion_x, mean_motion_y);
              Inc(bpos);
            end;
          ROQ_CODE_SLD :
            begin
              v4 := cin.V4Codebook[buf[bpos]];
              Inc(bpos);
     	      apply_vector_4x4(cin, xp,   yp,   @cin.V2CodeBook[v4.v2_index[0]]);
	      apply_vector_4x4(cin, xp+4, yp,   @cin.V2CodeBook[v4.v2_index[1]]);
	      apply_vector_4x4(cin, xp,   yp+4, @cin.V2CodeBook[v4.v2_index[2]]);
	      apply_vector_4x4(cin, xp+4, yp+4, @cin.V2CodeBook[v4.v2_index[3]]);
	    end;
	  ROQ_CODE_CCC :
	    begin
              for k:=0 to 3 do
              begin
	        x := xp; y := yp;
	        if (k and $01)<>0 then Inc(x, 4);
	        if (k and $02)<>0 then Inc(y, 4);

	        if (vqflg_pos < 0) then
	        begin
		  vqflg := PWord(@buf[bpos])^;
                  Inc(bpos, 2);
		  vqflg_pos := 7;
	        end;

                vqid := (vqflg shr (vqflg_pos * 2)) and $03;
	        Dec(vqflg_pos);

	        case vqid of
		  ROQ_CODE_MOT : begin end;
		  ROQ_CODE_FCC :
                    begin
		      apply_motion_4x4(cin, x, y, buf[bpos], mean_motion_x, mean_motion_y);
                      Inc(bpos);
		    end;
		  ROQ_CODE_SLD :
                    begin
                      v4 := cin.V4Codebook[buf[bpos]];
                      Inc(bpos);
		      apply_vector_2x2(cin, x,   y,   @cin.V2Codebook[v4.v2_index[0]]);
		      apply_vector_2x2(cin, x+2, y,   @cin.V2Codebook[v4.v2_index[1]]);
		      apply_vector_2x2(cin, x,   y+2, @cin.V2Codebook[v4.v2_index[2]]);
		      apply_vector_2x2(cin, x+2, y+2, @cin.V2Codebook[v4.v2_index[3]]);
		    end;
		  ROQ_CODE_CCC :
                    begin
                      apply_vector_2x2(cin, x,   y,   @cin.V2Codebook[buf[bpos+0]]);
		      apply_vector_2x2(cin, x+2, y,   @cin.V2Codebook[buf[bpos+1]]);
		      apply_vector_2x2(cin, x,   y+2, @cin.V2Codebook[buf[bpos+2]]);
		      apply_vector_2x2(cin, x+2, y+2, @cin.V2Codebook[buf[bpos+3]]);
		      Inc(bpos, 4);
		    end;
	        end;
              end;
            end;
        end;

        Inc(xp, 8);
      end;

      Inc(yp, 8);
    end;

    Inc(xpos, 16);
    if (xpos >= cin.Width) then
    begin
      Dec(xpos, cin.Width);
      Inc(ypos, 16);
    end;

    if (ypos >= cin.Height) then Break;
  end;

  if (cin.CurrentFrame = 0) then
    Move(cin.FrameBuffer[0]^, cin.FrameBuffer[1]^, cin.Width * cin.Height * 3)
  else
  begin
    tp := cin.FrameBuffer[0];
    cin.FrameBuffer[0] := cin.FrameBuffer[1];
    cin.FrameBuffer[1] := tp;
  end;

  FreeMem(buf);

  Inc(cin.CurrentFrame);
End;

procedure RoQ_ReadCodebook(cin: PCinematic);
var nv1, nv2: Cardinal;
begin
  nv1 := cin.Chunk.arg shr 8;
  nv2 := cin.Chunk.arg and $FF;
  if (nv1 = 0) then nv1 := 256;
  if (nv2 = 0) and ((nv1 * SizeOf(TRoQ_V2_Codebook)) < cin.Chunk.size) then nv2 := 256;

  cin.fp.Read(cin.V2CodeBook[0], SizeOf(TRoQ_V2_Codebook) * nv1);
  cin.fp.Read(cin.V4CodeBook[0], SizeOf(TRoQ_V4_Codebook) * nv2);
end;

function RoQ_ReadNextFrame(cin: PCinematic): Boolean;
begin
  Result := False;
  if cin.fp.Position >= cin.fp.Size then
    Exit;

  while cin.fp.Position < cin.fp.Size do
  begin
    RoQ_ReadChunk(cin);

    if cin.fp.Position >= cin.fp.Size then Exit;
    if cin.Chunk.size = 0 then Continue;

    case cin.Chunk.id of
      ROQ_SOUND_MONO,
      ROQ_SOUND_STEREO : RoQ_ReadAudio(cin);
      ROQ_QUAD_CODEBOOK: RoQ_ReadCodebook(cin);
      ROQ_QUAD_VQ: begin
                     RoQ_ReadVideo(cin);
                     Break;
                   end;  
    else
      RoQ_SkipChunk(cin);
    end;
  end;

  Result := True;
end;

end.
